from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from pathlib import Path

    from semantic_release.version.version import Version


class IVersionReplacer(metaclass=ABCMeta):
    """
    Interface for subclasses that replace a version string in a source file.

    Methods generally have a base implementation are implemented here but
    likely just provide a not-supported message but return gracefully

    This class cannot be instantiated directly but must be inherited from
    and implement the designated abstract methods.
    """

    @classmethod
    def __subclasshook__(cls, subclass: type) -> bool:
        # Validate that the subclass implements all of the abstract methods.
        # This supports isinstance and issubclass checks.
        return bool(
            cls is IVersionReplacer
            and all(
                bool(hasattr(subclass, method) and callable(getattr(subclass, method)))
                for method in IVersionReplacer.__abstractmethods__
            )
        )

    @abstractmethod
    def parse(self) -> set[Version]:
        """
        Return a set of the versions which can be parsed from the file.
        Because a source can match in multiple places, this method returns a
        set of matches. Generally, there should only be one element in this
        set (i.e. even if the version is specified in multiple places, it
        should be the same version in each place), but enforcing that condition
        is not mandatory or expected.
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def replace(self, new_version: Version) -> str:
        """
        Replace the version in the source content with `new_version`, and return
        the updated content.

        :param new_version: The new version number as a `Version` instance
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def update_file_w_version(
        self, new_version: Version, noop: bool = False
    ) -> Path | None:
        """
        This method reads the underlying file, replaces each occurrence of the
        matched pattern, then writes the updated file.

        :param new_version: The new version number as a `Version` instance
        """
        raise NotImplementedError  # pragma: no cover



import pytest

from semantic_release.commit_parser import (
    ConventionalCommitParser,
    ConventionalCommitParserOptions,
    EmojiCommitParser,
    EmojiParserOptions,
)

from tests.const import (
    CONVENTIONAL_COMMITS_CHORE,
    CONVENTIONAL_COMMITS_MAJOR,
    CONVENTIONAL_COMMITS_MINOR,
    CONVENTIONAL_COMMITS_PATCH,
    EMOJI_COMMITS_CHORE,
    EMOJI_COMMITS_MAJOR,
    EMOJI_COMMITS_MINOR,
    EMOJI_COMMITS_PATCH,
)

# Note scipy defined in ./scipy.py as already used there


@pytest.fixture(scope="session")
def default_conventional_parser() -> ConventionalCommitParser:
    return ConventionalCommitParser()


@pytest.fixture(scope="session")
def default_conventional_parser_options(
    default_conventional_parser: ConventionalCommitParser,
) -> ConventionalCommitParserOptions:
    return default_conventional_parser.get_default_options()


@pytest.fixture(scope="session")
def default_emoji_parser() -> EmojiCommitParser:
    return EmojiCommitParser()


@pytest.fixture(scope="session")
def default_emoji_parser_options(
    default_emoji_parser: EmojiCommitParser,
) -> EmojiParserOptions:
    return default_emoji_parser.get_default_options()


@pytest.fixture(scope="session")
def conventional_major_commits():
    return CONVENTIONAL_COMMITS_MAJOR


@pytest.fixture(scope="session")
def conventional_minor_commits():
    return CONVENTIONAL_COMMITS_MINOR


@pytest.fixture(scope="session")
def conventional_patch_commits():
    return CONVENTIONAL_COMMITS_PATCH


@pytest.fixture(scope="session")
def conventional_chore_commits():
    return CONVENTIONAL_COMMITS_CHORE


@pytest.fixture(scope="session")
def emoji_major_commits():
    return EMOJI_COMMITS_MAJOR


@pytest.fixture(scope="session")
def emoji_minor_commits():
    return EMOJI_COMMITS_MINOR


@pytest.fixture(scope="session")
def emoji_patch_commits():
    return EMOJI_COMMITS_PATCH


@pytest.fixture(scope="session")
def emoji_chore_commits():
    return EMOJI_COMMITS_CHORE



from requests import PreparedRequest
from requests.auth import AuthBase


class TokenAuth(AuthBase):
    """
    requests Authentication for token based authorization.
    This allows us to attach the Authorization header with
    a token to a session.
    """

    def __init__(self, token: str) -> None:
        self.token = token

    def __eq__(self, other: object) -> bool:
        return self.token == getattr(other, "token", None)

    def __ne__(self, other: object) -> bool:
        return not self == other

    def __call__(self, req: PreparedRequest) -> PreparedRequest:
        req.headers["Authorization"] = f"token {self.token}"
        return req



from __future__ import annotations

import os
import subprocess
import sys
from collections import defaultdict
from datetime import datetime, timezone
from typing import TYPE_CHECKING

import click
import shellingham  # type: ignore[import]
from click_option_group import MutuallyExclusiveOptionGroup, optgroup
from git import GitCommandError, Repo
from requests import HTTPError

from semantic_release.changelog.release_history import ReleaseHistory
from semantic_release.cli.changelog_writer import (
    generate_release_notes,
    write_changelog_files,
)
from semantic_release.cli.github_actions_output import (
    PersistenceMode,
    VersionGitHubActionsOutput,
)
from semantic_release.cli.util import noop_report, rprint
from semantic_release.const import DEFAULT_SHELL, DEFAULT_VERSION
from semantic_release.enums import LevelBump
from semantic_release.errors import (
    BuildDistributionsError,
    DetachedHeadGitError,
    GitCommitEmptyIndexError,
    GitFetchError,
    InternalError,
    LocalGitError,
    UnexpectedResponse,
    UnknownUpstreamBranchError,
    UpstreamBranchChangedError,
)
from semantic_release.gitproject import GitProject
from semantic_release.globals import logger
from semantic_release.hvcs.github import Github
from semantic_release.hvcs.remote_hvcs_base import RemoteHvcsBase
from semantic_release.version.algorithm import (
    next_version,
    tags_and_versions,
)
from semantic_release.version.translator import VersionTranslator

if TYPE_CHECKING:  # pragma: no cover
    from pathlib import Path
    from typing import Mapping, Sequence

    from git.refs.tag import Tag

    from semantic_release.cli.cli_context import CliContextObj
    from semantic_release.version.declaration import IVersionReplacer
    from semantic_release.version.version import Version


def is_forced_prerelease(
    as_prerelease: bool, forced_level_bump: LevelBump | None, prerelease: bool
) -> bool:
    """
    Determine if this release is forced to have prerelease on/off.
    If ``force_prerelease`` is set then yes.
    Otherwise if we are forcing a specific level bump without force_prerelease,
    it's False.
    Otherwise (``force_level is None``) use the value of ``prerelease``
    """
    local_vars = list(locals().items())
    logger.debug(
        "%s: %s",
        is_forced_prerelease.__name__,
        str.join(", ", iter(f"{k} = {v}" for k, v in local_vars)),
    )
    return (
        as_prerelease
        or forced_level_bump is LevelBump.PRERELEASE_REVISION
        or ((forced_level_bump is None) and prerelease)
    )


def last_released(
    repo_dir: Path, tag_format: str, add_partial_tags: bool = False
) -> tuple[Tag, Version] | None:
    with Repo(str(repo_dir)) as git_repo:
        ts_and_vs = tags_and_versions(
            git_repo.tags,
            VersionTranslator(tag_format=tag_format, add_partial_tags=add_partial_tags),
        )

    return ts_and_vs[0] if ts_and_vs else None


def version_from_forced_level(
    repo_dir: Path, forced_level_bump: LevelBump, translator: VersionTranslator
) -> Version:
    with Repo(str(repo_dir)) as git_repo:
        ts_and_vs = tags_and_versions(git_repo.tags, translator)

    # If we have no tags, return the default version
    if not ts_and_vs:
        # Since the translator is configured by the user, we can't guarantee that it will
        # be able to parse the default version. So we first cast it to a tag using the default
        # value and the users configured tag format, then parse it back to a version object
        default_initial_version = translator.from_tag(
            translator.str_to_tag(DEFAULT_VERSION)
        )
        if default_initial_version is None:
            # This should never happen, but if it does, it's a bug
            raise InternalError(
                "Translator was unable to parse the embedded default version"
            )
        return default_initial_version.bump(forced_level_bump)

    _, latest_version = ts_and_vs[0]
    if forced_level_bump is not LevelBump.PRERELEASE_REVISION:
        return latest_version.bump(forced_level_bump)

    # We need to find the latest version with the prerelease token
    # we're looking for, and return that version + an increment to
    # the prerelease revision.

    # NOTE this can probably be cleaned up.
    # ts_and_vs are in order, so check if we're looking at prereleases
    # for the same (major, minor, patch) as the latest version.
    # If we are, we can increment the revision and we're done. If
    # we don't find a prerelease targeting this version with the same
    # token as the one we're looking to prerelease, we can use revision 1.
    for _, version in ts_and_vs:
        if not (
            version.major == latest_version.major
            and version.minor == latest_version.minor
            and version.patch == latest_version.patch
        ):
            break
        if (
            version.is_prerelease
            and version.prerelease_token == translator.prerelease_token
        ):
            return version.bump(LevelBump.PRERELEASE_REVISION)
    return latest_version.to_prerelease(token=translator.prerelease_token, revision=1)


def apply_version_to_source_files(
    repo_dir: Path,
    version_declarations: Sequence[IVersionReplacer],
    version: Version,
    noop: bool = False,
) -> list[str]:
    if len(version_declarations) < 1:
        return []

    if not noop:
        logger.debug("Updating version %s in repository files...", version)

    paths = [
        decl.update_file_w_version(new_version=version, noop=noop)
        for decl in version_declarations
    ]

    repo_filepaths = [
        str(updated_file.relative_to(repo_dir))
        for updated_file in paths
        if updated_file is not None
    ]

    if noop:
        noop_report(
            str.join(
                "",
                [
                    "would have updated versions in the following paths:",
                    *[f"\n    {filepath}" for filepath in repo_filepaths],
                ],
            )
        )

    return repo_filepaths


def shell(
    cmd: str, *, env: Mapping[str, str] | None = None, check: bool = True
) -> subprocess.CompletedProcess:
    shell: str | None
    try:
        shell, _ = shellingham.detect_shell()
    except shellingham.ShellDetectionFailure:
        logger.warning("failed to detect shell, using default shell: %s", DEFAULT_SHELL)
        logger.debug("stack trace", exc_info=True)
        shell = DEFAULT_SHELL

    if not shell:
        raise TypeError("'shell' is None")

    shell_cmd_param = defaultdict(
        lambda: "-c",
        {
            "cmd": "/c",
            "powershell": "-Command",
            "pwsh": "-Command",
        },
    )

    return subprocess.run(  # noqa: S603
        [shell, shell_cmd_param[shell], cmd],
        env=(env or {}),
        check=check,
    )


def is_windows() -> bool:
    return sys.platform == "win32"


def get_windows_env() -> Mapping[str, str | None]:
    return {
        environment_variable: os.getenv(environment_variable, None)
        for environment_variable in (
            "ALLUSERSAPPDATA",
            "ALLUSERSPROFILE",
            "APPDATA",
            "COMMONPROGRAMFILES",
            "COMMONPROGRAMFILES(X86)",
            "DEFAULTUSERPROFILE",
            "HOMEPATH",
            "PATHEXT",
            "PROFILESFOLDER",
            "PROGRAMFILES",
            "PROGRAMFILES(X86)",
            "SYSTEM",
            "SYSTEM16",
            "SYSTEM32",
            "SYSTEMDRIVE",
            "SYSTEMPROFILE",
            "SYSTEMROOT",
            "TEMP",
            "TMP",
            "USERNAME",  # must include for python getpass.getuser() on windows
            "USERPROFILE",
            "USERSID",
            "WINDIR",
        )
    }


def build_distributions(
    build_command: str | None,
    build_command_env: Mapping[str, str] | None = None,
    noop: bool = False,
) -> None:
    """
    Run the build command to build the distributions.

    :param build_command: The build command to run.
    :param build_command_env: The environment variables to use when running the
        build command.
    :param noop: Whether or not to run the build command.

    :raises: BuildDistributionsError: if the build command fails
    """
    if not build_command:
        rprint("[green]No build command specified, skipping")
        return

    if noop:
        noop_report(f"would have run the build_command {build_command}")
        return

    logger.info("Running build command %s", build_command)
    rprint(f"[bold green]:hammer_and_wrench: Running build command: {build_command}")

    build_env_vars: dict[str, str] = dict(
        filter(
            lambda k_v: k_v[1] is not None,  # type: ignore[arg-type]
            {
                # Common values
                "PATH": os.getenv("PATH", ""),
                "HOME": os.getenv("HOME", None),
                "VIRTUAL_ENV": os.getenv("VIRTUAL_ENV", None),
                # Windows environment variables
                **(get_windows_env() if is_windows() else {}),
                # affects build decisions
                "CI": os.getenv("CI", None),
                # Identifies which CI environment
                "GITHUB_ACTIONS": os.getenv("GITHUB_ACTIONS", None),
                "GITLAB_CI": os.getenv("GITLAB_CI", None),
                "GITEA_ACTIONS": os.getenv("GITEA_ACTIONS", None),
                "BITBUCKET_CI": (
                    str(True).lower()
                    if os.getenv("BITBUCKET_REPO_FULL_NAME", None)
                    else None
                ),
                "PSR_DOCKER_GITHUB_ACTION": os.getenv("PSR_DOCKER_GITHUB_ACTION", None),
                **(build_command_env or {}),
            }.items(),
        )
    )

    try:
        shell(build_command, env=build_env_vars, check=True)
        rprint("[bold green]Build completed successfully!")
    except subprocess.CalledProcessError as exc:
        logger.exception(exc)
        logger.error("Build command failed with exit code %s", exc.returncode)  # noqa: TRY400
        raise BuildDistributionsError from exc


@click.command(
    short_help="Detect and apply a new version",
    context_settings={
        "help_option_names": ["-h", "--help"],
    },
)
@optgroup.group("Print flags", cls=MutuallyExclusiveOptionGroup)
@optgroup.option(
    "--print", "print_only", is_flag=True, help="Print the next version and exit"
)
@optgroup.option(
    "--print-tag",
    "print_only_tag",
    is_flag=True,
    help="Print the next version tag and exit",
)
@optgroup.option(
    "--print-last-released",
    is_flag=True,
    help="Print the last released version and exit",
)
@optgroup.option(
    "--print-last-released-tag",
    is_flag=True,
    help="Print the last released version tag and exit",
)
@click.option(
    "--as-prerelease",
    "as_prerelease",
    is_flag=True,
    help="Ensure the next version to be released is a prerelease version",
)
@click.option(
    "--prerelease-token",
    "prerelease_token",
    default=None,
    help="Force the next version to use this prerelease token, if it is a prerelease",
)
@click.option(
    "--major",
    "force_level",
    flag_value="major",
    help="Force the next version to be a major release",
)
@click.option(
    "--minor",
    "force_level",
    flag_value="minor",
    help="Force the next version to be a minor release",
)
@click.option(
    "--patch",
    "force_level",
    flag_value="patch",
    help="Force the next version to be a patch release",
)
@click.option(
    "--prerelease",
    "force_level",
    flag_value="prerelease_revision",
    help="Force the next version to be a prerelease",
)
@click.option(
    "--commit/--no-commit",
    "commit_changes",
    default=True,
    help="Whether or not to commit changes locally",
)
@click.option(
    "--tag/--no-tag",
    "create_tag",
    default=True,
    help="Whether or not to create a tag for the new version",
)
@click.option(
    "--changelog/--no-changelog",
    "update_changelog",
    default=True,
    help="Whether or not to update the changelog",
)
@click.option(
    "--push/--no-push",
    "push_changes",
    default=True,
    help="Whether or not to push the new commit and tag to the remote",
)
@click.option(
    "--vcs-release/--no-vcs-release",
    "make_vcs_release",
    default=True,
    help="Whether or not to create a release in the remote VCS, if supported",
)
@click.option(
    "--build-metadata",
    "build_metadata",
    default=os.getenv("PSR_BUILD_METADATA"),
    help="Build metadata to append to the new version",
)
@click.option(
    "--skip-build",
    "skip_build",
    default=False,
    is_flag=True,
    help="Skip building the current project",
)
@click.pass_obj
def version(  # noqa: C901
    cli_ctx: CliContextObj,
    print_only: bool,
    print_only_tag: bool,
    print_last_released: bool,
    print_last_released_tag: bool,
    as_prerelease: bool,
    prerelease_token: str | None,
    commit_changes: bool,
    create_tag: bool,
    update_changelog: bool,
    push_changes: bool,
    make_vcs_release: bool,
    build_metadata: str | None,
    skip_build: bool,
    force_level: str | None = None,
) -> None:
    """
    Detect the semantically correct next version that should be applied to your
    project.

    By default:

    * Write this new version to the project metadata locations specified
    in the configuration file

    * Create a new commit with these locations and any other assets configured
    to be included in a release

    * Tag this commit according the configured format, with a tag that uniquely
    identifies the version being released.

    * Push the new tag and commit to the remote for the repository

    * Create a release (if supported) in the remote VCS for this tag
    """
    ctx = click.get_current_context()

    # Enable any cli overrides of configuration before asking for the runtime context
    config = cli_ctx.raw_config

    # We can short circuit updating the release if we are only printing the last released version
    if print_last_released or print_last_released_tag:
        # TODO: get tag format a better way
        if not (
            last_release := last_released(
                config.repo_dir,
                tag_format=config.tag_format,
                add_partial_tags=config.add_partial_tags,
            )
        ):
            logger.warning("No release tags found.")
            return

        click.echo(last_release[0] if print_last_released_tag else last_release[1])
        return

    # TODO: figure out --print of next version with & without branch validation
    # do you always need a prerelease token if its not --as-prerelease?
    runtime = cli_ctx.runtime_ctx
    translator = runtime.version_translator

    parser = runtime.commit_parser
    hvcs_client = runtime.hvcs_client
    assets = runtime.assets
    commit_author = runtime.commit_author
    commit_message = runtime.commit_message
    major_on_zero = runtime.major_on_zero
    no_verify = runtime.no_git_verify
    opts = runtime.global_cli_options
    add_partial_tags = config.add_partial_tags
    gha_output = VersionGitHubActionsOutput(
        gh_client=hvcs_client if isinstance(hvcs_client, Github) else None,
        mode=(
            PersistenceMode.TEMPORARY
            if opts.noop or (not commit_changes and not create_tag)
            else PersistenceMode.PERMANENT
        ),
        released=False,
    )

    forced_level_bump = None if not force_level else LevelBump.from_string(force_level)
    prerelease = is_forced_prerelease(
        as_prerelease=as_prerelease,
        forced_level_bump=forced_level_bump,
        prerelease=runtime.prerelease,
    )

    if prerelease_token:
        logger.info("Forcing use of %s as the prerelease token", prerelease_token)
        translator.prerelease_token = prerelease_token

    # Check if the repository is shallow and unshallow it if necessary
    # This ensures we have the full history for commit analysis
    project = GitProject(
        directory=runtime.repo_dir,
        commit_author=runtime.commit_author,
        credential_masker=runtime.masker,
    )
    if project.is_shallow_clone():
        logger.info("Repository is a shallow clone, converting to full clone...")
        project.git_unshallow(noop=opts.noop)

    # Only push if we're committing changes
    if push_changes and not commit_changes and not create_tag:
        logger.info("changes will not be pushed because --no-commit disables pushing")
        push_changes &= commit_changes

    # Only push if we're creating a tag
    if push_changes and not create_tag and not commit_changes:
        logger.info("new tag will not be pushed because --no-tag disables pushing")
        push_changes &= create_tag

    # Only make a release if we're pushing the changes
    if make_vcs_release and not push_changes:
        logger.info(
            "No vcs release will be created because pushing changes is disabled"
        )
        make_vcs_release &= push_changes

    if not forced_level_bump:
        with Repo(str(runtime.repo_dir)) as git_repo:
            new_version = next_version(
                repo=git_repo,
                translator=translator,
                commit_parser=parser,
                prerelease=prerelease,
                major_on_zero=major_on_zero,
                allow_zero_version=runtime.allow_zero_version,
            )
    else:
        logger.warning(
            "Forcing a '%s' release due to '--%s' command-line flag",
            force_level,
            (
                force_level
                if forced_level_bump is not LevelBump.PRERELEASE_REVISION
                else "prerelease"
            ),
        )

        new_version = version_from_forced_level(
            repo_dir=runtime.repo_dir,
            forced_level_bump=forced_level_bump,
            translator=translator,
        )

        # We only turn the forced version into a prerelease if the user has specified
        # that that is what they want on the command-line; otherwise we assume they are
        # forcing a full release
        new_version = (
            new_version.to_prerelease(token=translator.prerelease_token)
            if prerelease
            else new_version.finalize_version()
        )

    if build_metadata:
        new_version.build_metadata = build_metadata

    # Update GitHub Actions output value with new version & set delayed write
    gha_output.version = new_version
    if isinstance(hvcs_client, Github):
        ctx.call_on_close(gha_output.write_if_possible)

    # Make string variant of version or appropriate tag as necessary
    version_to_print = str(new_version) if not print_only_tag else new_version.as_tag()

    # Print the new version so that command-line output capture will work
    click.echo(version_to_print)

    with Repo(str(runtime.repo_dir)) as git_repo:
        # TODO: performance improvement - cache the result of tags_and_versions (previously done in next_version())
        previously_released_versions = {
            v for _, v in tags_and_versions(git_repo.tags, translator)
        }

    # If the new version has already been released, we fail and abort if strict;
    # otherwise we exit with 0.
    if new_version in previously_released_versions:
        err_msg = str.join(
            " ",
            [
                "[bold orange1]No release will be made,",
                f"{new_version!s} has already been released!",
            ],
        )

        if opts.strict:
            click.echo(err_msg, err=True)
            ctx.exit(2)

        rprint(err_msg)
        return

    if print_only or print_only_tag:
        return

    # TODO: need a better way as this is inconsistent if releasing older version patches
    if last_release := last_released(config.repo_dir, tag_format=config.tag_format):
        # If we have a last release, we can set the previous version for the
        # GitHub Actions output
        gha_output.prev_version = last_release[1]

    with Repo(str(runtime.repo_dir)) as git_repo:
        release_history = ReleaseHistory.from_git_history(
            repo=git_repo,
            translator=translator,
            commit_parser=parser,
            exclude_commit_patterns=runtime.changelog_excluded_commit_patterns,
        )

    rprint(f"[bold green]The next version is: [white]{new_version!s}[/white]! :rocket:")

    commit_date = datetime.now(timezone.utc).astimezone()  # Locale-aware timestamp
    try:
        # Create release object for the new version
        # This will be used to generate the changelog prior to the commit and/or tag
        release_history = release_history.release(
            new_version,
            tagger=commit_author,
            committer=commit_author,
            tagged_date=commit_date,
        )
    except ValueError as ve:
        click.echo(str(ve), err=True)
        ctx.exit(1)

    all_paths_to_add: list[str] = []

    if update_changelog:
        # Write changelog files & add them to the list of files to commit
        all_paths_to_add.extend(
            write_changelog_files(
                runtime_ctx=runtime,
                release_history=release_history,
                hvcs_client=hvcs_client,
                noop=opts.noop,
            )
        )

    # Apply the new version to the source files
    files_with_new_version_written = apply_version_to_source_files(
        repo_dir=runtime.repo_dir,
        version_declarations=runtime.version_declarations,
        version=new_version,
        noop=opts.noop,
    )
    all_paths_to_add.extend(files_with_new_version_written)
    all_paths_to_add.extend(assets or [])

    # Build distributions before committing any changes - this way if the
    # build fails, modifications to the source code won't be committed
    if skip_build:
        rprint("[bold orange1]Skipping build due to --skip-build flag")
    else:
        try:
            build_distributions(
                build_command=runtime.build_command,
                build_command_env={
                    # User defined overrides of environment (from config)
                    **runtime.build_command_env,
                    # PSR injected environment variables
                    "NEW_VERSION": str(new_version),
                    "PACKAGE_NAME": runtime.project_metadata.get("name", ""),
                },
                noop=opts.noop,
            )
        except BuildDistributionsError as exc:
            click.echo(str(exc), err=True)
            click.echo("Build failed, aborting release", err=True)
            ctx.exit(1)

    license_cfg = runtime.project_metadata.get(
        "license-expression",
        runtime.project_metadata.get(
            "license",
            "",
        ),
    )

    license_cfg = "" if not isinstance(license_cfg, (str, dict)) else license_cfg
    license_cfg = (
        license_cfg.get("text", "") if isinstance(license_cfg, dict) else license_cfg
    )

    gha_output.release_notes = release_notes = generate_release_notes(
        hvcs_client,
        release=release_history.released[new_version],
        template_dir=runtime.template_dir,
        history=release_history,
        style=runtime.changelog_style,
        mask_initial_release=runtime.changelog_mask_initial_release,
        license_name="" if not isinstance(license_cfg, str) else license_cfg,
    )

    # Preparing for committing changes; we always stage files even if we're not committing them in order to support a two-stage commit
    project.git_add(paths=all_paths_to_add, noop=opts.noop)
    if commit_changes:
        # NOTE: If we haven't modified any source code then we skip trying to make a commit
        # and any tag that we apply will be to the HEAD commit (made outside of
        # running PSR
        try:
            project.git_commit(
                message=commit_message.format(version=new_version),
                date=int(commit_date.timestamp()),
                no_verify=no_verify,
                noop=opts.noop,
            )
        except GitCommitEmptyIndexError:
            logger.info("No local changes to add to any commit, skipping")
            commit_changes = False

    # Tag the version after potentially creating a new HEAD commit.
    # This way if no source code is modified, i.e. all metadata updates
    # are disabled, and the changelog generation is disabled or it's not
    # modified, then the HEAD commit will be tagged as a release commit
    # despite not being made by PSR
    if create_tag:
        project.git_tag(
            tag_name=new_version.as_tag(),
            message=new_version.as_tag(),
            isotimestamp=commit_date.isoformat(),
            noop=opts.noop,
        )

        with Repo(str(runtime.repo_dir)) as git_repo:
            gha_output.commit_sha = git_repo.head.commit.hexsha

    if push_changes:
        remote_url = runtime.hvcs_client.remote_url(
            use_token=not runtime.ignore_token_for_push
        )

        if commit_changes:
            # Verify that the upstream branch has not changed before pushing
            # This prevents conflicts if another commit was pushed while we were preparing the release
            # We check HEAD~1 because we just made a release commit
            try:
                project.verify_upstream_unchanged(
                    local_ref="HEAD~1",
                    upstream_ref=config.remote.name,
                    remote_url=remote_url,
                    noop=opts.noop,
                )
            except UpstreamBranchChangedError as exc:
                click.echo(str(exc), err=True)
                click.echo(
                    "Upstream branch has changed. Please pull the latest changes and try again.",
                    err=True,
                )
                ctx.exit(1)
            except (
                DetachedHeadGitError,
                GitCommandError,
                UnknownUpstreamBranchError,
                GitFetchError,
                LocalGitError,
            ) as exc:
                click.echo(str(exc), err=True)
                click.echo("Unable to verify upstream due to error!", err=True)
                ctx.exit(1)

            # TODO: integrate into push branch
            with Repo(str(runtime.repo_dir)) as git_repo:
                active_branch = git_repo.active_branch.name

            project.git_push_branch(
                remote_url=remote_url,
                branch=active_branch,
                noop=opts.noop,
            )

        if create_tag:
            # push specific tag refspec (that we made) to remote
            project.git_push_tag(
                remote_url=remote_url,
                tag=new_version.as_tag(),
                noop=opts.noop,
            )
            # Create or update partial tags for releases
            if add_partial_tags and not prerelease:
                partial_tags = [new_version.as_major_tag(), new_version.as_minor_tag()]
                # If build metadata is set, also retag the version without the metadata
                if build_metadata:
                    partial_tags.append(new_version.as_patch_tag())

                for partial_tag in partial_tags:
                    project.git_tag(
                        tag_name=partial_tag,
                        message=f"{partial_tag} => {new_version.as_tag()}",
                        isotimestamp=commit_date.isoformat(),
                        noop=opts.noop,
                        force=True,
                    )
                    project.git_push_tag(
                        remote_url=remote_url,
                        tag=partial_tag,
                        noop=opts.noop,
                        force=True,
                    )

    # Update GitHub Actions output value now that release has occurred
    gha_output.released = True

    if not make_vcs_release:
        return

    if not isinstance(hvcs_client, RemoteHvcsBase):
        logger.info("Remote does not support releases. Skipping release creation...")
        return

    exception: Exception | None = None
    help_message = ""
    try:
        hvcs_client.create_release(
            tag=new_version.as_tag(),
            release_notes=release_notes,
            prerelease=new_version.is_prerelease,
            assets=assets,
            noop=opts.noop,
        )
    except HTTPError as err:
        exception = err
    except UnexpectedResponse as err:
        exception = err
        help_message = str.join(
            " ",
            [
                "Before re-running, make sure to clean up any artifacts",
                "on the hvcs that may have already been created.",
            ],
        )
        help_message = str.join(
            "\n",
            [
                "Unexpected response from remote VCS!",
                help_message,
            ],
        )
    except Exception as err:  # noqa: BLE001
        # TODO: Remove this catch-all exception handler in the future
        exception = err
    finally:
        if exception is not None:
            logger.exception(exception)
            click.echo(str(exception), err=True)
            if help_message:
                click.echo(help_message, err=True)
            click.echo(
                f"Failed to create release on {hvcs_client.__class__.__name__}!",
                err=True,
            )
            ctx.exit(1)



from __future__ import annotations

import os
from datetime import timezone
from typing import TYPE_CHECKING, cast

import pytest
from freezegun import freeze_time
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from tests.const import EXAMPLE_PROJECT_LICENSE, MAIN_PROG_NAME, VERSION_SUBCMD
from tests.fixtures.repos import (
    repo_w_git_flow_w_alpha_prereleases_n_conventional_commits,
)
from tests.util import actions_output_to_dict, assert_successful_exit_code

if TYPE_CHECKING:
    from semantic_release.hvcs.github import Github

    from tests.conftest import GetStableDateNowFn, RunCliFn
    from tests.fixtures.example_project import ExProjectDir
    from tests.fixtures.git_repo import (
        BuiltRepoResult,
        GenerateDefaultReleaseNotesFromDefFn,
        GetCfgValueFromDefFn,
        GetHvcsClientFromRepoDefFn,
        GetVersionsFromRepoBuildDefFn,
        SplitRepoActionsByReleaseTagsFn,
    )


@pytest.mark.parametrize(
    "repo_result",
    [lazy_fixture(repo_w_git_flow_w_alpha_prereleases_n_conventional_commits.__name__)],
)
def test_version_writes_github_actions_output(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
    example_project_dir: ExProjectDir,
    get_cfg_value_from_def: GetCfgValueFromDefFn,
    get_hvcs_client_from_repo_def: GetHvcsClientFromRepoDefFn,
    generate_default_release_notes_from_def: GenerateDefaultReleaseNotesFromDefFn,
    split_repo_actions_by_release_tags: SplitRepoActionsByReleaseTagsFn,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    stable_now_date: GetStableDateNowFn,
):
    mock_output_file = example_project_dir / "action.out"
    repo_def = repo_result["definition"]
    tag_format_str = cast(str, get_cfg_value_from_def(repo_def, "tag_format_str"))
    all_versions = get_versions_from_repo_build_def(repo_def)
    latest_release_version = all_versions[-1]
    release_tag = tag_format_str.format(version=latest_release_version)
    previous_version = all_versions[-2] if len(all_versions) > 1 else None
    hvcs_client = cast("Github", get_hvcs_client_from_repo_def(repo_def))
    repo_actions_per_version = split_repo_actions_by_release_tags(
        repo_definition=repo_def
    )
    expected_gha_output = {
        "released": str(True).lower(),
        "version": latest_release_version,
        "tag": release_tag,
        "link": hvcs_client.create_release_url(release_tag),
        "commit_sha": "0" * 40,
        "is_prerelease": str(latest_release_version.is_prerelease).lower(),
        "previous_version": str(previous_version) if previous_version else "",
        "release_notes": generate_default_release_notes_from_def(
            version_actions=repo_actions_per_version[latest_release_version],
            hvcs=hvcs_client,
            previous_version=previous_version,
            license_name=EXAMPLE_PROJECT_LICENSE,
            mask_initial_release=get_cfg_value_from_def(
                repo_def, "mask_initial_release"
            ),
        ),
    }

    # Remove the previous tag & version commit
    repo_result["repo"].git.tag(release_tag, delete=True)
    repo_result["repo"].git.reset("HEAD~1", hard=True)

    # Act
    with freeze_time(stable_now_date().astimezone(timezone.utc)):
        cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--no-push"]
        result = run_cli(
            cli_cmd[1:], env={"GITHUB_OUTPUT": str(mock_output_file.resolve())}
        )

    assert_successful_exit_code(result, cli_cmd)

    # Update the expected output with the commit SHA
    expected_gha_output["commit_sha"] = repo_result["repo"].head.commit.hexsha

    if not mock_output_file.exists():
        pytest.fail(
            f"Expected output file {mock_output_file} to be created, but it does not exist."
        )

    # Extract the output
    with open(mock_output_file, encoding="utf-8", newline=os.linesep) as rfd:
        action_outputs = actions_output_to_dict(rfd.read())

    # Evaluate
    expected_keys = set(expected_gha_output.keys())
    actual_keys = set(action_outputs.keys())
    key_difference = expected_keys.symmetric_difference(actual_keys)

    assert not key_difference, f"Unexpected keys found: {key_difference}"

    assert expected_gha_output["released"] == action_outputs["released"]
    assert expected_gha_output["version"] == action_outputs["version"]
    assert expected_gha_output["tag"] == action_outputs["tag"]
    assert expected_gha_output["is_prerelease"] == action_outputs["is_prerelease"]
    assert expected_gha_output["link"] == action_outputs["link"]
    assert expected_gha_output["previous_version"] == action_outputs["previous_version"]
    assert expected_gha_output["commit_sha"] == action_outputs["commit_sha"]
    assert expected_gha_output["release_notes"] == action_outputs["release_notes"]



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock

import pytest
from git import Commit, Repo, TagReference

from semantic_release.enums import LevelBump
from semantic_release.version.algorithm import (
    _increment_version,
    _traverse_graph_for_commits,
    tags_and_versions,
)
from semantic_release.version.translator import VersionTranslator
from semantic_release.version.version import Version

from tests.fixtures.repos import repo_w_initial_commit

if TYPE_CHECKING:
    from typing import Sequence


@pytest.mark.usefixtures(repo_w_initial_commit.__name__)
def test_traverse_graph_for_commits():
    # Setup fake git graph
    """
    * merge commit 6 (start) [3636363]
    |\
    | * commit 5 [3535353]
    | * commit 4 [3434343]
    |/
    * commit 3 [3333333]
    * commit 2 [3232323]
    * commit 1 [3131313]
    * v1.0.0 [3030303]
    """
    repo = Repo()
    v1_commit = Commit(repo, binsha=b"0" * 20, parents=[])

    class TagReferenceOverride(TagReference):
        commit = v1_commit  # mocking the commit property

    v1_tag = TagReferenceOverride(repo, "refs/tags/v1.0.0", check_path=False)

    trunk = Commit(
        repo,
        binsha=b"3" * 20,
        parents=[
            Commit(
                repo,
                binsha=b"2" * 20,
                parents=[
                    Commit(repo, binsha=b"1" * 20, parents=[v1_commit]),
                ],
            ),
        ],
    )
    start_commit = Commit(
        repo,
        binsha=b"6" * 20,
        parents=[
            trunk,
            Commit(
                repo,
                binsha=b"5" * 20,
                parents=[
                    Commit(repo, binsha=b"4" * 20, parents=[trunk]),
                ],
            ),
        ],
    )

    commit_1 = trunk.parents[0].parents[0]
    commit_2 = trunk.parents[0]
    commit_3 = trunk
    commit_4 = start_commit.parents[1].parents[0]
    commit_5 = start_commit.parents[1]
    commit_6 = start_commit

    expected_commit_order = [
        commit_6.hexsha,
        commit_5.hexsha,
        commit_4.hexsha,
        commit_3.hexsha,
        commit_2.hexsha,
        commit_1.hexsha,
    ]

    # Execute
    with mock.patch.object(
        repo, repo.iter_commits.__name__, return_value=iter([v1_commit])
    ):
        actual_commit_order = [
            commit.hexsha
            for commit in _traverse_graph_for_commits(
                head_commit=start_commit,
                latest_release_tag_str=v1_tag.name,
            )
        ]

    # Verify
    assert expected_commit_order == actual_commit_order


@pytest.mark.parametrize(
    "tags, sorted_tags",
    [
        (
            ["v1.0.0", "v1.1.0", "v1.1.1"],
            ["v1.1.1", "v1.1.0", "v1.0.0"],
        ),
        (
            ["v1.1.0", "v1.0.0", "v1.1.1"],
            ["v1.1.1", "v1.1.0", "v1.0.0"],
        ),
        (
            ["v1.1.1", "v1.1.0", "v1.0.0"],
            ["v1.1.1", "v1.1.0", "v1.0.0"],
        ),
        # Examples from https://semver.org/#spec-item-11 (or inspired, where not all
        # version structures are supported)
        (
            ["v1.0.0", "v2.0.0", "v2.1.1", "v2.1.0"],
            ["v2.1.1", "v2.1.0", "v2.0.0", "v1.0.0"],
        ),
        (
            [
                "v1.0.0-rc.1",
                "v1.0.0-beta.2",
                "v1.0.0-beta.11",
                "v1.0.0-alpha.1",
                "v1.0.0-alpha.beta.1",
                "v1.0.0",
            ],
            [
                "v1.0.0",
                "v1.0.0-rc.1",
                "v1.0.0-beta.11",
                "v1.0.0-beta.2",
                "v1.0.0-alpha.beta.1",
                "v1.0.0-alpha.1",
            ],
        ),
    ],
)
def test_sorted_repo_tags_and_versions(tags: list[str], sorted_tags: list[str]):
    repo = Repo()
    translator = VersionTranslator()
    tagrefs = [repo.tag(tag) for tag in tags]
    actual = [t.name for t, _ in tags_and_versions(tagrefs, translator)]
    assert sorted_tags == actual


@pytest.mark.parametrize(
    "tag_format, invalid_tags, valid_tags",
    [
        pytest.param(
            tag_format,
            invalid_tags,
            valid_tags,
            id=test_id,
        )
        for test_id, tag_format, invalid_tags, valid_tags in [
            (
                "traditional-v-prefixed-versions",
                "v{version}",
                (
                    "0.3",  # no v-prefix
                    "test-v1.1.0",  # extra prefix
                    "v1.1.0-test-test",  # bad suffix
                ),
                [
                    "v1.0.0-rc.1",
                    "v1.0.0-beta.2",
                    "v1.0.0-beta.11",
                    "v1.0.0-alpha.1",
                    "v1.0.0-alpha.beta.1",
                    "v1.0.0",
                ],
            ),
            (
                "monorepo-style-versions",
                "pkg1-v{version}",
                (
                    "0.3",  # no pkg or version prefix
                    "v1.1.0",  # no pkg prefix
                    "pkg1-v1.1.0-test-test",  # bad suffix
                    "pkg2-v1.1.0",  # wrong package prefix
                ),
                [
                    "pkg1-v1.0.0-rc.1",
                    "pkg1-v1.0.0-beta.2",
                    "pkg1-v1.0.0-beta.11",
                    "pkg1-v1.0.0-alpha.1",
                    "pkg1-v1.0.0-alpha.beta.1",
                    "pkg1-v1.0.0",
                ],
            ),
        ]
    ],
)
def test_tags_and_versions_ignores_invalid_tags_as_versions(
    tag_format: str,
    invalid_tags: Sequence[str],
    valid_tags: Sequence[str],
):
    repo = Repo()
    translator = VersionTranslator(tag_format=tag_format)
    tagrefs = [repo.tag(tag) for tag in (*valid_tags, *invalid_tags)]
    actual = [t.name for t, _ in tags_and_versions(tagrefs, translator)]
    assert set(valid_tags) == set(actual)


@pytest.mark.parametrize(
    str.join(
        ", ",
        [
            "latest_version",
            "latest_full_version",
            "level_bump",
            "prerelease",
            "prerelease_token",
            "expected_version",
        ],
    ),
    [
        # NOTE: level_bump != LevelBump.NO_RELEASE, we return early in the
        # algorithm to discount this case
        # NOTE: you can only perform a PRERELEASE_REVISION bump on a previously
        # prerelease version and if you are requesting a prerelease
        (
            "1.0.1-rc.1",
            "1.0.0",
            LevelBump.PRERELEASE_REVISION,
            True,
            "rc",
            "1.0.1-rc.2",
        ),
        *[
            (
                "1.0.0",
                "1.0.0",
                bump_level,
                prerelease,
                "rc",
                expected_version,
            )
            for bump_level, prerelease, expected_version in [
                (LevelBump.PATCH, False, "1.0.1"),
                (LevelBump.PATCH, True, "1.0.1-rc.1"),
                (LevelBump.MINOR, False, "1.1.0"),
                (LevelBump.MINOR, True, "1.1.0-rc.1"),
                (LevelBump.MAJOR, False, "2.0.0"),
                (LevelBump.MAJOR, True, "2.0.0-rc.1"),
            ]
        ],
        (
            "1.2.4-rc.1",
            "1.2.3",
            LevelBump.PRERELEASE_REVISION,
            True,
            "rc",
            "1.2.4-rc.2",
        ),
        *[
            (
                "1.2.4-rc.1",
                "1.2.3",
                bump_level,
                prerelease,
                "rc",
                expected_version,
            )
            for bump_level, prerelease, expected_version in [
                (LevelBump.PATCH, False, "1.2.4"),
                (LevelBump.PATCH, True, "1.2.4-rc.2"),
                (LevelBump.MINOR, False, "1.3.0"),
                (LevelBump.MINOR, True, "1.3.0-rc.1"),
                (LevelBump.MAJOR, False, "2.0.0"),
                (LevelBump.MAJOR, True, "2.0.0-rc.1"),
            ]
        ],
        (
            "2.0.0-rc.1",
            "1.19.3",
            LevelBump.PRERELEASE_REVISION,
            True,
            "rc",
            "2.0.0-rc.2",
        ),
        *[
            (
                "2.0.0-rc.1",
                "1.22.0",
                bump_level,
                prerelease,
                "rc",
                expected_version,
            )
            for bump_level, prerelease, expected_version in [
                (LevelBump.PATCH, False, "2.0.0"),
                (LevelBump.PATCH, True, "2.0.0-rc.2"),
                (LevelBump.MINOR, False, "2.0.0"),
                (LevelBump.MINOR, True, "2.0.0-rc.2"),
                (LevelBump.MAJOR, False, "2.0.0"),
                (LevelBump.MAJOR, True, "2.0.0-rc.2"),
            ]
        ],
    ],
)
def test_increment_version_no_major_on_zero(
    latest_version: str,
    latest_full_version: str,
    level_bump: LevelBump,
    prerelease: bool,
    prerelease_token: str,
    expected_version: str,
):
    actual = _increment_version(
        latest_version=Version.parse(latest_version),
        latest_full_version=Version.parse(latest_full_version),
        level_bump=level_bump,
        prerelease=prerelease,
        prerelease_token=prerelease_token,
        major_on_zero=False,
        allow_zero_version=True,
    )
    assert expected_version == str(actual)


@pytest.mark.parametrize(
    "latest_version, latest_full_version, level_bump, prerelease, prerelease_token",
    [
        # NOTE: level_bump != LevelBump.NO_RELEASE, we return early in the
        # algorithm to discount this case
        # NOTE: you can only perform a PRERELEASE_REVISION bump on a previously
        # prerelease version and if you are requesting a prerelease
        (
            "1.0.0",
            "1.0.0",
            LevelBump.PRERELEASE_REVISION,
            False,
            "rc",
        ),
        (
            "1.0.0",
            "1.0.0",
            LevelBump.PRERELEASE_REVISION,
            True,
            "rc",
        ),
    ],
)
def test_increment_version_invalid_operation(
    latest_version: str,
    latest_full_version: str,
    level_bump: LevelBump,
    prerelease: bool,
    prerelease_token: str,
):
    with pytest.raises(ValueError):
        _increment_version(
            latest_version=Version.parse(latest_version),
            latest_full_version=Version.parse(latest_full_version),
            level_bump=level_bump,
            prerelease=prerelease,
            prerelease_token=prerelease_token,
            major_on_zero=False,
            allow_zero_version=True,
        )



"""Legacy commit parser from Python Semantic Release 1.0"""

from __future__ import annotations

import re

from git.objects.commit import Commit
from pydantic.dataclasses import dataclass

from semantic_release.commit_parser._base import CommitParser, ParserOptions
from semantic_release.commit_parser.token import ParsedCommit, ParseError, ParseResult
from semantic_release.commit_parser.util import breaking_re, parse_paragraphs
from semantic_release.enums import LevelBump
from semantic_release.globals import logger

re_parser = re.compile(r"(?P<subject>[^\n]+)" + r"(:?\n\n(?P<text>.+))?", re.DOTALL)


@dataclass
class TagParserOptions(ParserOptions):
    minor_tag: str = ":sparkles:"
    patch_tag: str = ":nut_and_bolt:"


def _logged_parse_error(commit: Commit, error: str) -> ParseError:
    logger.debug(error)
    return ParseError(commit, error=error)


class TagCommitParser(CommitParser[ParseResult, TagParserOptions]):
    """
    Parse a commit message according to the 1.0 version of python-semantic-release.
    It expects a tag of some sort in the commit message and will use the rest of the
    first line as changelog content.
    """

    # TODO: Deprecate in lieu of get_default_options()
    parser_options = TagParserOptions

    @staticmethod
    def get_default_options() -> TagParserOptions:
        return TagParserOptions()

    def parse(self, commit: Commit) -> ParseResult | list[ParseResult]:
        message = str(commit.message)

        # Attempt to parse the commit message with a regular expression
        parsed = re_parser.match(message)
        if not parsed:
            return _logged_parse_error(
                commit, error=f"Unable to parse the given commit message: {message!r}"
            )

        subject = parsed.group("subject")

        # Check tags for minor or patch
        if self.options.minor_tag in message:
            level = "feature"
            level_bump = LevelBump.MINOR
            if subject:
                subject = subject.replace(self.options.minor_tag, "")

        elif self.options.patch_tag in message:
            level = "fix"
            level_bump = LevelBump.PATCH
            if subject:
                subject = subject.replace(self.options.patch_tag, "")

        else:
            # We did not find any tags in the commit message
            return _logged_parse_error(
                commit, error=f"Unable to parse the given commit message: {message!r}"
            )

        if parsed.group("text"):
            descriptions = parse_paragraphs(parsed.group("text"))
        else:
            descriptions = []
        descriptions.insert(0, subject.strip())

        # Look for descriptions of breaking changes
        breaking_descriptions = [
            match.group(1)
            for match in (breaking_re.match(p) for p in descriptions[1:])
            if match
        ]
        if breaking_descriptions:
            level = "breaking"
            level_bump = LevelBump.MAJOR
            logger.debug(
                "commit %s upgraded to a %s level_bump due to included breaking descriptions",
                commit.hexsha[:8],
                level_bump,
            )

        logger.debug(
            "commit %s introduces a %s level_bump", commit.hexsha[:8], level_bump
        )

        return ParsedCommit(
            bump=level_bump,
            type=level,
            scope="",
            descriptions=descriptions,
            breaking_descriptions=breaking_descriptions,
            commit=commit,
        )



"""Utilities for command-line functionality"""

from __future__ import annotations

import json
import sys
from pathlib import Path
from textwrap import dedent, indent
from typing import Any

import rich
import rich.markup
import tomlkit
from tomlkit.exceptions import TOMLKitError

from semantic_release.errors import InvalidConfiguration
from semantic_release.globals import logger


def rprint(msg: str) -> None:
    """Rich-prints to stderr so that redirection of command output isn't cluttered"""
    rich.print(msg, file=sys.stderr)


def noop_report(msg: str) -> None:
    """
    Rich-prints a msg with a standard prefix to report when an action is not being
    taken due to a "noop" flag
    """
    rprint(f"[bold cyan][:shield: NOP] {rich.markup.escape(msg)}")


def indented(msg: str, prefix: str = " " * 4) -> str:
    """
    Convenience function for text-formatting for the console.

    Ensures the least indented line of the msg string is indented by ``prefix`` with
    consistent alignment of the remainder of ``msg`` irrespective of the level of
    indentation in the Python source code
    """
    return indent(dedent(msg), prefix=prefix)


def parse_toml(raw_text: str) -> dict[Any, Any]:
    """
    Attempts to parse raw configuration for semantic_release
    using tomlkit.loads, raising InvalidConfiguration if the
    TOML is invalid or there's no top level "semantic_release"
    or "tool.semantic_release" keys
    """
    try:
        toml_text = tomlkit.loads(raw_text).unwrap()
    except TOMLKitError as exc:
        raise InvalidConfiguration(str(exc)) from exc

    # Look for [tool.semantic_release]
    cfg_text = toml_text.get("tool", {}).get("semantic_release")
    if cfg_text is not None:
        return cfg_text
    # Look for [semantic_release] or return {} if not found
    return toml_text.get("semantic_release", {})


def load_raw_config_file(config_file: Path | str) -> dict[Any, Any]:
    """
    Load raw configuration as a dict from the filename specified
    by config_filename, trying the following parsing methods:

    1. try to parse with tomli.load (guessing it's a TOML file)
    2. try to parse with json.load (guessing it's a JSON file)
    3. raise InvalidConfiguration if none of the above parsing
       methods work

    This function will also raise FileNotFoundError if it is raised
    while trying to read the specified configuration file
    """
    logger.info("Loading configuration from %s", config_file)
    raw_text = (Path() / config_file).resolve().read_text(encoding="utf-8")
    try:
        logger.debug("Trying to parse configuration %s in TOML format", config_file)
        return parse_toml(raw_text)
    except InvalidConfiguration as e:
        logger.debug("Configuration %s is invalid TOML: %s", config_file, str(e))
        logger.debug("trying to parse %s as JSON", config_file)
        try:
            # could be a "parse_json" function but it's a one-liner here
            return json.loads(raw_text)["semantic_release"]
        except KeyError:
            # valid configuration, but no "semantic_release" or "tool.semantic_release"
            # top level key
            logger.debug(
                "configuration has no 'semantic_release' or 'tool.semantic_release' "
                "top-level key"
            )
            return {}
        except json.JSONDecodeError as jde:
            raise InvalidConfiguration(
                dedent(
                    f"""
                    None of the supported configuration parsers were able to parse
                    the configuration file {config_file}:
                    * TOML: {e!s}
                    * JSON: {jde!s}
                    """
                )
            ) from jde



from __future__ import annotations

import os
from textwrap import dedent
from typing import TYPE_CHECKING
from unittest import mock

import pytest

from semantic_release.cli.github_actions_output import VersionGitHubActionsOutput
from semantic_release.hvcs.github import Github
from semantic_release.version.version import Version

from tests.const import EXAMPLE_HVCS_DOMAIN, EXAMPLE_REPO_NAME, EXAMPLE_REPO_OWNER
from tests.util import actions_output_to_dict

if TYPE_CHECKING:
    from pathlib import Path


BASE_VCS_URL = f"https://{EXAMPLE_HVCS_DOMAIN}/{EXAMPLE_REPO_OWNER}/{EXAMPLE_REPO_NAME}"


@pytest.mark.parametrize(
    "prev_version, version, released, is_prerelease",
    [
        ("1.2.2", "1.2.3", True, False),
        ("1.2.2", "1.2.3-alpha.1", True, True),
        ("1.2.2", "1.2.2", False, False),
        ("1.2.2-alpha.1", "1.2.2-alpha.1", False, True),
        (None, "1.2.3", True, False),
    ],
)
def test_version_github_actions_output_format(
    released: bool, version: str, is_prerelease: bool, prev_version: str
):
    commit_sha = "0" * 40  # 40 zeroes to simulate a SHA-1 hash
    release_notes = dedent(
        """\
        ## Changes
        - Added new feature
        - Fixed bug
        """
    )
    expected_output = (
        dedent(
            f"""\
            released={'true' if released else 'false'}
            version={version}
            tag=v{version}
            is_prerelease={'true' if is_prerelease else 'false'}
            link={BASE_VCS_URL}/releases/tag/v{version}
            previous_version={prev_version or ""}
            commit_sha={commit_sha}
            """
        )
        + f"release_notes<<EOF{os.linesep}{release_notes}EOF{os.linesep}"
    )

    with mock.patch.dict(os.environ, {}, clear=True):
        actual_output_text = VersionGitHubActionsOutput(
            gh_client=Github(f"{BASE_VCS_URL}.git", hvcs_domain=EXAMPLE_HVCS_DOMAIN),
            released=released,
            version=Version.parse(version),
            commit_sha=commit_sha,
            release_notes=release_notes,
            prev_version=Version.parse(prev_version) if prev_version else None,
        ).to_output_text()

    # Evaluate (expected -> actual)
    assert expected_output == actual_output_text


def test_version_github_actions_output_fails_if_missing_released_param():
    output = VersionGitHubActionsOutput(
        gh_client=Github(f"{BASE_VCS_URL}.git"),
        version=Version.parse("1.2.3"),
    )

    # Execute with expected failure
    with pytest.raises(ValueError, match="required outputs were not set"):
        output.to_output_text()


def test_version_github_actions_output_fails_if_missing_commit_sha_param():
    output = VersionGitHubActionsOutput(
        gh_client=Github(f"{BASE_VCS_URL}.git"),
        released=True,
        version=Version.parse("1.2.3"),
    )

    # Execute with expected failure
    with pytest.raises(ValueError, match="required outputs were not set"):
        output.to_output_text()


def test_version_github_actions_output_fails_if_missing_release_notes_param():
    output = VersionGitHubActionsOutput(
        gh_client=Github(f"{BASE_VCS_URL}.git"),
        released=True,
        version=Version.parse("1.2.3"),
    )

    # Execute with expected failure
    with pytest.raises(ValueError, match="required outputs were not set"):
        output.to_output_text()


def test_version_github_actions_output_writes_to_github_output_if_available(
    tmp_path: Path,
):
    mock_output_file = tmp_path / "action.out"
    prev_version_str = "1.2.2"
    version_str = "1.2.3"
    commit_sha = "0" * 40  # 40 zeroes to simulate a SHA-1 hash
    release_notes = dedent(
        """\
        ## Changes
        - Added new feature
        - Fixed bug
        """
    )

    patched_environ = {"GITHUB_OUTPUT": str(mock_output_file.resolve())}

    with mock.patch.dict(os.environ, patched_environ, clear=True):
        VersionGitHubActionsOutput(
            gh_client=Github(f"{BASE_VCS_URL}.git", hvcs_domain=EXAMPLE_HVCS_DOMAIN),
            version=Version.parse(version_str),
            released=True,
            commit_sha=commit_sha,
            release_notes=release_notes,
            prev_version=Version.parse(prev_version_str),
        ).write_if_possible()

    with open(mock_output_file, encoding="utf-8", newline=os.linesep) as rfd:
        action_outputs = actions_output_to_dict(rfd.read())

    # Evaluate (expected -> actual)
    assert version_str == action_outputs["version"]
    assert str(True).lower() == action_outputs["released"]
    assert str(False).lower() == action_outputs["is_prerelease"]
    assert f"{BASE_VCS_URL}/releases/tag/v{version_str}" == action_outputs["link"]
    assert f"v{version_str}" == action_outputs["tag"]
    assert commit_sha == action_outputs["commit_sha"]
    assert prev_version_str == action_outputs["previous_version"]
    assert release_notes == action_outputs["release_notes"]


def test_version_github_actions_output_no_error_if_not_in_gha(
    monkeypatch: pytest.MonkeyPatch,
):
    output = VersionGitHubActionsOutput(
        gh_client=Github(f"{BASE_VCS_URL}.git"),
        version=Version.parse("1.2.3"),
        released=True,
        commit_sha="0" * 40,  # 40 zeroes to simulate a SHA-1 hash
    )

    monkeypatch.delenv("GITHUB_OUTPUT", raising=False)
    output.write_if_possible()



import os
import sys
from datetime import datetime, timezone

sys.path.insert(0, os.path.abspath("."))
sys.path.insert(0, os.path.abspath(".."))

import semantic_release  # noqa: E402

author_name = "Python Semantic Release Team"

# -- General configuration ------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinxcontrib.apidoc",
]

autodoc_default_options = {"ignore-module-all": True}

templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"
project = "python-semantic-release"
current_year = datetime.now(timezone.utc).astimezone().year
copyright = f"{current_year}, {author_name}"  # noqa: A001

version = semantic_release.__version__
release = semantic_release.__version__


exclude_patterns = ["_build"]
pygments_style = "sphinx"
html_theme = "furo"
htmlhelp_basename = "python-semantic-releasedoc"


# -- Automatically run sphinx-apidoc --------------------------------------

docs_path = os.path.dirname(__file__)
apidoc_output_dir = os.path.join(docs_path, "api", "modules")
apidoc_module_dir = os.path.join(docs_path, "..", "src")
apidoc_separate_modules = True
apidoc_module_first = True
apidoc_extra_args = ["-d", "3"]


def setup(app):  # type: ignore[no-untyped-def]  # noqa: ARG001,ANN001,ANN201
    pass


# -- Options for LaTeX output ---------------------------------------------
latex_documents = [
    (
        "index",
        "python-semantic-release.tex",
        "python-semantic-release Documentation",
        author_name,
        "manual",
    ),
]


# -- Options for manual page output ---------------------------------------
man_pages = [
    (
        "index",
        "python-semantic-release",
        "python-semantic-release Documentation",
        [author_name],
        1,
    )
]


# -- Options for Texinfo output -------------------------------------------
texinfo_documents = [
    (
        "index",
        "python-semantic-release",
        "python-semantic-release Documentation",
        author_name,
        "python-semantic-release",
        "One line description of project.",
        "Miscellaneous",
    ),
]


# -- Options for Epub output ----------------------------------------------

# Bibliographic Dublin Core info.
epub_title = "python-semantic-release"
epub_author = author_name
epub_publisher = author_name
epub_copyright = copyright
epub_exclude_files = ["search.html"]



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from semantic_release.hvcs.github import Github

from tests.const import MAIN_PROG_NAME, VERSION_SUBCMD
from tests.fixtures.repos import repo_w_trunk_only_conventional_commits
from tests.util import assert_exit_code

if TYPE_CHECKING:
    from unittest.mock import MagicMock

    from requests_mock import Mocker

    from tests.conftest import RunCliFn
    from tests.e2e.conftest import StripLoggingMessagesFn
    from tests.fixtures.git_repo import BuiltRepoResult, GetVersionsFromRepoBuildDefFn


@pytest.mark.parametrize(
    "repo_result",
    [lazy_fixture(repo_w_trunk_only_conventional_commits.__name__)],
)
def test_version_already_released_when_strict(
    repo_result: BuiltRepoResult,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    run_cli: RunCliFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    strip_logging_messages: StripLoggingMessagesFn,
):
    """
    Given repo has no new changes since the last release,
    When running the version command in strict mode,
    Then no version release should happen, which means no code changes, no build, no commit,
    no tag, no push, and no vcs release creation while returning an exit code of 2.
    """
    repo = repo_result["repo"]
    latest_release_version = get_versions_from_repo_build_def(
        repo_result["definition"]
    )[-1]
    expected_error_msg = f"[bold orange1]No release will be made, {latest_release_version} has already been released!"

    # Setup: take measurement before running the version command
    repo_status_before = repo.git.status(short=True)
    head_before = repo.head.commit.hexsha
    tags_before = sorted([tag.name for tag in repo.tags])

    # Act
    cli_cmd = [MAIN_PROG_NAME, "--strict", VERSION_SUBCMD]
    result = run_cli(cli_cmd[1:], env={Github.DEFAULT_ENV_TOKEN_NAME: "1234"})

    # take measurement after running the version command
    repo_status_after = repo.git.status(short=True)
    head_after = repo.head.commit.hexsha
    tags_after = sorted([tag.name for tag in repo.tags])

    # Evaluate
    assert_exit_code(2, result, cli_cmd)
    assert f"{latest_release_version}\n" == result.stdout
    assert f"{expected_error_msg}\n" == strip_logging_messages(result.stderr)

    # assert nothing else happened (no code changes, no commit, no tag, no push, no vcs release)
    assert repo_status_before == repo_status_after
    assert head_before == head_after
    assert tags_before == tags_after
    assert mocked_git_push.call_count == 0
    assert post_mocker.call_count == 0


@pytest.mark.parametrize(
    "repo_result", [lazy_fixture(repo_w_trunk_only_conventional_commits.__name__)]
)
def test_version_on_nonrelease_branch_when_strict(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    strip_logging_messages: StripLoggingMessagesFn,
):
    """
    Given repo is on a non-release branch,
    When running the version command in strict mode,
    Then no version release should happen which means no code changes, no build, no commit,
    no tag, no push, and no vcs release creation while returning an exit code of 2.
    """
    repo = repo_result["repo"]

    # Setup
    branch = repo.create_head("next").checkout()
    expected_error_msg = (
        f"branch '{branch.name}' isn't in any release groups; no release will be made\n"
    )
    repo_status_before = repo.git.status(short=True)
    head_before = repo.head.commit.hexsha
    tags_before = sorted([tag.name for tag in repo.tags])

    # Act
    cli_cmd = [MAIN_PROG_NAME, "--strict", VERSION_SUBCMD]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_exit_code(2, result, cli_cmd)
    assert not result.stdout
    assert expected_error_msg == strip_logging_messages(result.stderr)

    # assert nothing else happened (no code changes, no commit, no tag, no push, no vcs release)
    tags_after = sorted([tag.name for tag in repo.tags])
    assert repo_status_before == repo.git.status(short=True)
    assert head_before == repo.head.commit.hexsha
    assert tags_before == tags_after
    assert mocked_git_push.call_count == 0
    assert post_mocker.call_count == 0



"""Entrypoint for the `semantic-release` module."""
# ruff: noqa: T201, print statements are fine here as this is for cli entry only

from __future__ import annotations

import sys
from traceback import format_exception

from semantic_release import globals
from semantic_release.cli.commands.main import main as cli_main
from semantic_release.enums import SemanticReleaseLogLevels


def main() -> None:
    try:
        cli_main(args=sys.argv[1:])
        print("semantic-release completed successfully.", file=sys.stderr)
    except KeyboardInterrupt:
        print("\n-- User Abort! --", file=sys.stderr)
        sys.exit(127)
    except Exception as err:  # noqa: BLE001, graceful error handling across application
        if globals.log_level <= SemanticReleaseLogLevels.DEBUG:
            print(f"{err.__class__.__name__}: {err}\n", file=sys.stderr)
            etype, value, traceback = sys.exc_info()
            print(
                str.join(
                    "",
                    format_exception(
                        etype,
                        value,
                        traceback,
                        limit=None,
                        chain=True,
                    )[:-1],
                ),
                file=sys.stderr,
            )

        print(
            str.join("\n", [f"::ERROR:: {line}" for line in str(err).splitlines()]),
            file=sys.stderr,
        )

        if globals.log_level > SemanticReleaseLogLevels.DEBUG:
            print(
                "Run semantic-release in very verbose mode (-vv) to see the full traceback.",
                file=sys.stderr,
            )

        sys.exit(1)


if __name__ == "__main__":
    main()



"""Module for git related operations."""

from __future__ import annotations

from contextlib import nullcontext
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING

from git import GitCommandError, Repo

from semantic_release.cli.masking_filter import MaskingFilter
from semantic_release.cli.util import indented, noop_report
from semantic_release.errors import (
    DetachedHeadGitError,
    GitAddError,
    GitCommitEmptyIndexError,
    GitCommitError,
    GitFetchError,
    GitPushError,
    GitTagError,
    LocalGitError,
    UnknownUpstreamBranchError,
    UpstreamBranchChangedError,
)
from semantic_release.globals import logger

if TYPE_CHECKING:  # pragma: no cover
    from contextlib import _GeneratorContextManager
    from logging import Logger
    from typing import Sequence

    from git import Actor


class GitProject:
    def __init__(
        self,
        directory: Path | str = ".",
        commit_author: Actor | None = None,
        credential_masker: MaskingFilter | None = None,
    ) -> None:
        self._project_root = Path(directory).resolve()
        self._logger = logger
        self._cred_masker = credential_masker or MaskingFilter()
        self._commit_author = commit_author

    @property
    def project_root(self) -> Path:
        return self._project_root

    @property
    def logger(self) -> Logger:
        return self._logger

    def _get_custom_environment(
        self,
        repo: Repo,
        custom_vars: dict[str, str] | None = None,
    ) -> nullcontext[None] | _GeneratorContextManager[None]:
        """
        git.custom_environment is a context manager but
        is not reentrant, so once we have "used" it
        we need to throw it away and re-create it in
        order to use it again
        """
        author_vars = (
            {
                "GIT_AUTHOR_NAME": self._commit_author.name,
                "GIT_AUTHOR_EMAIL": self._commit_author.email,
                "GIT_COMMITTER_NAME": self._commit_author.name,
                "GIT_COMMITTER_EMAIL": self._commit_author.email,
            }
            if self._commit_author
            else {}
        )

        custom_env_vars = {
            **author_vars,
            **(custom_vars or {}),
        }

        return (
            nullcontext()
            if not custom_env_vars
            else repo.git.custom_environment(**custom_env_vars)
        )

    def is_dirty(self) -> bool:
        with Repo(str(self.project_root)) as repo:
            return repo.is_dirty()

    def is_shallow_clone(self) -> bool:
        """
        Check if the repository is a shallow clone.

        :return: True if the repository is a shallow clone, False otherwise
        """
        with Repo(str(self.project_root)) as repo:
            shallow_file = Path(repo.git_dir, "shallow")
            return shallow_file.exists()

    def git_unshallow(self, noop: bool = False) -> None:
        """
        Convert a shallow clone to a full clone by fetching the full history.

        :param noop: Whether or not to actually run the unshallow command
        """
        if noop:
            noop_report("would have run:\n" "    git fetch --unshallow")
            return

        with Repo(str(self.project_root)) as repo:
            try:
                self.logger.info("Converting shallow clone to full clone...")
                repo.git.fetch("--unshallow")
                self.logger.info("Repository unshallowed successfully")
            except GitCommandError as err:
                # If the repository is already a full clone, git fetch --unshallow will fail
                # with "fatal: --unshallow on a complete repository does not make sense"
                # We can safely ignore this error by checking the stderr message
                stderr = str(err.stderr) if err.stderr else ""
                if "does not make sense" in stderr or "complete repository" in stderr:
                    self.logger.debug("Repository is already a full clone")
                else:
                    self.logger.exception(str(err))
                    raise

    def git_add(
        self,
        paths: Sequence[Path | str],
        force: bool = False,
        strict: bool = False,
        noop: bool = False,
    ) -> None:
        if noop:
            noop_report(
                indented(
                    f"""\
                    would have run:
                        git add {str.join(" ", [str(Path(p)) for p in paths])}
                    """
                )
            )
            return

        git_args = dict(
            filter(
                lambda k_v: k_v[1],  # if truthy
                {
                    "force": force,
                }.items(),
            )
        )

        with Repo(str(self.project_root)) as repo:
            # TODO: in future this loop should be 1 line:
            # repo.index.add(all_paths_to_add, force=False)  # noqa: ERA001
            # but since 'force' is deliberately ineffective (as in docstring) in gitpython 3.1.18
            # we have to do manually add each filepath, and catch the exception if it is an ignored file
            for updated_path in paths:
                try:
                    repo.git.add(str(Path(updated_path)), **git_args)
                except GitCommandError as err:  # noqa: PERF203, acceptable performance loss
                    err_msg = f"Failed to add path ({updated_path}) to index"
                    if strict:
                        self.logger.exception(str(err))
                        raise GitAddError(err_msg) from err
                    self.logger.warning(err_msg)

    def git_commit(
        self,
        message: str,
        date: int | None = None,
        commit_all: bool = False,
        no_verify: bool = False,
        noop: bool = False,
    ) -> None:
        git_args = dict(
            filter(
                lambda k_v: k_v[1],  # if truthy
                {
                    "a": commit_all,
                    "m": message,
                    "date": date,
                    "no_verify": no_verify,
                }.items(),
            )
        )

        if noop:
            command = (
                f"""\
                GIT_AUTHOR_NAME={self._commit_author.name} \\
                GIT_AUTHOR_EMAIL={self._commit_author.email} \\
                GIT_COMMITTER_NAME={self._commit_author.name} \\
                GIT_COMMITTER_EMAIL={self._commit_author.email} \\
                """
                if self._commit_author
                else ""
            )

            # Indents the newlines so that terminal formatting is happy - note the
            # git commit line of the output is 24 spaces indented too
            # Only this message needs such special handling because of the newlines
            # that might be in a commit message between the subject and body
            indented_commit_message = message.replace("\n\n", "\n\n" + " " * 24)

            command += f"git commit -m '{indented_commit_message}'"
            command += "--all" if commit_all else ""
            command += "--no-verify" if no_verify else ""

            noop_report(
                indented(
                    f"""\
                    would have run:
                        {command}
                    """
                )
            )
            return

        with Repo(str(self.project_root)) as repo:
            has_index_changes = bool(repo.index.diff("HEAD"))
            has_working_changes = self.is_dirty()
            will_commit_files = has_index_changes or (
                has_working_changes and commit_all
            )

            if not will_commit_files:
                raise GitCommitEmptyIndexError("No changes to commit!")

            with self._get_custom_environment(repo):
                try:
                    repo.git.commit(**git_args)
                except GitCommandError as err:
                    self.logger.exception(str(err))
                    raise GitCommitError("Failed to commit changes") from err

    def git_tag(
        self,
        tag_name: str,
        message: str,
        isotimestamp: str,
        force: bool = False,
        noop: bool = False,
    ) -> None:
        try:
            datetime.fromisoformat(isotimestamp)
        except ValueError as err:
            raise ValueError("Invalid timestamp format") from err

        if noop:
            command = str.join(
                " ",
                filter(
                    None,
                    [
                        f"GIT_COMMITTER_DATE={isotimestamp}",
                        *(
                            [
                                f"GIT_AUTHOR_NAME={self._commit_author.name}",
                                f"GIT_AUTHOR_EMAIL={self._commit_author.email}",
                                f"GIT_COMMITTER_NAME={self._commit_author.name}",
                                f"GIT_COMMITTER_EMAIL={self._commit_author.email}",
                            ]
                            if self._commit_author
                            else [""]
                        ),
                        f"git tag -a {tag_name} -m '{message}'",
                        "--force" if force else "",
                    ],
                ),
            ).strip()

            noop_report(
                indented(
                    f"""\
                    would have run:
                        {command}
                    """
                )
            )
            return

        with Repo(str(self.project_root)) as repo, self._get_custom_environment(
            repo,
            {"GIT_COMMITTER_DATE": isotimestamp},
        ):
            try:
                repo.git.tag(tag_name, a=True, m=message, force=force)
            except GitCommandError as err:
                self.logger.exception(str(err))
                raise GitTagError(f"Failed to create tag ({tag_name})") from err

    def git_push_branch(self, remote_url: str, branch: str, noop: bool = False) -> None:
        if noop:
            noop_report(
                indented(
                    f"""\
                    would have run:
                        git push {self._cred_masker.mask(remote_url)} {branch}
                    """
                )
            )
            return

        with Repo(str(self.project_root)) as repo:
            try:
                repo.git.push(remote_url, branch)
            except GitCommandError as err:
                self.logger.exception(str(err))
                raise GitPushError(
                    f"Failed to push branch ({branch}) to remote"
                ) from err

    def git_push_tag(
        self, remote_url: str, tag: str, noop: bool = False, force: bool = False
    ) -> None:
        if noop:
            noop_report(
                indented(
                    f"""\
                    would have run:
                        git push {self._cred_masker.mask(remote_url)} tag {tag} {"--force" if force else ""}
                    """  # noqa: E501
                )
            )
            return

        with Repo(str(self.project_root)) as repo:
            try:
                repo.git.push(remote_url, "tag", tag, force=force)
            except GitCommandError as err:
                self.logger.exception(str(err))
                raise GitPushError(f"Failed to push tag ({tag}) to remote") from err

    def verify_upstream_unchanged(  # noqa: C901
        self,
        local_ref: str = "HEAD",
        upstream_ref: str = "origin",
        remote_url: str | None = None,
        noop: bool = False,
    ) -> None:
        """
        Verify that the upstream branch has not changed since the given local reference.

        :param local_ref: The local reference to compare against upstream (default: HEAD)
        :param upstream_ref: The name of the upstream remote or specific remote branch (default: origin)
        :param remote_url: Optional authenticated remote URL to use for fetching (default: None, uses configured remote)
        :param noop: Whether to skip the actual verification (for dry-run mode)

        :raises UpstreamBranchChangedError: If the upstream branch has changed
        """
        if not local_ref.strip():
            raise ValueError("Local reference cannot be empty")
        if not upstream_ref.strip():
            raise ValueError("Upstream reference cannot be empty")

        if noop:
            noop_report(
                indented(
                    """\
                    would have verified that upstream branch has not changed
                    """
                )
            )
            return

        with Repo(str(self.project_root)) as repo:
            # Get the current active branch
            try:
                active_branch = repo.active_branch
            except TypeError:
                # When in detached HEAD state, active_branch raises TypeError
                err_msg = (
                    "Repository is in detached HEAD state, cannot verify upstream state"
                )
                raise DetachedHeadGitError(err_msg) from None

            # Get the tracking branch (upstream branch)
            if (tracking_branch := active_branch.tracking_branch()) is not None:
                upstream_full_ref_name = tracking_branch.name
                self.logger.info("Upstream branch name: %s", upstream_full_ref_name)
            else:
                # If no tracking branch is set, derive it
                upstream_name = (
                    upstream_ref.strip()
                    if upstream_ref.find("/") == -1
                    else upstream_ref.strip().split("/", maxsplit=1)[0]
                )

                if not repo.remotes or upstream_name not in repo.remotes:
                    err_msg = "No remote found; cannot verify upstream state!"
                    raise UnknownUpstreamBranchError(err_msg)

                upstream_full_ref_name = (
                    f"{upstream_name}/{active_branch.name}"
                    if upstream_ref.find("/") == -1
                    else upstream_ref.strip()
                )

                if upstream_full_ref_name not in repo.refs:
                    err_msg = f"No upstream branch found for '{active_branch.name}'; cannot verify upstream state!"
                    raise UnknownUpstreamBranchError(err_msg)

            # Extract the remote name from the tracking branch
            # tracking_branch.name is in the format "remote/branch"
            remote_name, remote_branch_name = upstream_full_ref_name.split(
                "/", maxsplit=1
            )
            remote_ref_obj = repo.remotes[remote_name]

            # Fetch the latest changes from the remote
            self.logger.info("Fetching latest changes from remote '%s'", remote_name)
            try:
                # Check if we should use authenticated URL for fetch
                # Only use remote_url if:
                # 1. It's provided and different from the configured remote URL
                # 2. It contains authentication credentials (@ symbol)
                # 3. The configured remote is NOT a local path, file:// URL, or test URL (example.com)
                #    This ensures we don't break tests or local development
                configured_url = remote_ref_obj.url
                is_local_or_test_remote = (
                    configured_url.startswith(("file://", "/", "C:/", "H:/"))
                    or "example.com" in configured_url
                    or not configured_url.startswith(
                        (
                            "https://",
                            "http://",
                            "git://",
                            "git@",
                            "ssh://",
                            "git+ssh://",
                        )
                    )
                )

                use_authenticated_fetch = (
                    remote_url
                    and "@" in remote_url
                    and remote_url != configured_url
                    and not is_local_or_test_remote
                )

                if use_authenticated_fetch:
                    # Use authenticated remote URL for fetch
                    # Fetch the remote branch and update the local tracking ref
                    repo.git.fetch(
                        remote_url,
                        f"refs/heads/{remote_branch_name}:refs/remotes/{upstream_full_ref_name}",
                    )
                else:
                    # Use the default remote configuration for local paths,
                    # file:// URLs, test URLs, or when no authentication is needed
                    remote_ref_obj.fetch()
            except GitCommandError as err:
                self.logger.exception(str(err))
                err_msg = f"Failed to fetch from remote '{remote_name}'"
                raise GitFetchError(err_msg) from err

            # Get the SHA of the upstream branch
            try:
                upstream_commit_ref = remote_ref_obj.refs[remote_branch_name].commit
                upstream_sha = upstream_commit_ref.hexsha
            except AttributeError as err:
                self.logger.exception(str(err))
                err_msg = f"Unable to determine upstream branch SHA for '{upstream_full_ref_name}'"
                raise GitFetchError(err_msg) from err

            # Get the SHA of the specified ref (default: HEAD)
            try:
                local_commit = repo.commit(repo.git.rev_parse(local_ref))
            except GitCommandError as err:
                self.logger.exception(str(err))
                err_msg = f"Unable to determine the SHA for local ref '{local_ref}'"
                raise LocalGitError(err_msg) from err

            # Compare the two SHAs
            if local_commit.hexsha != upstream_sha and not any(
                commit.hexsha == upstream_sha for commit in local_commit.iter_parents()
            ):
                err_msg = str.join(
                    "\n",
                    (
                        f"[LOCAL SHA] {local_commit.hexsha} != {upstream_sha} [UPSTREAM SHA].",
                        f"Upstream branch '{upstream_full_ref_name}' has changed!",
                    ),
                )
                raise UpstreamBranchChangedError(err_msg)

            self.logger.info(
                "Verified upstream branch '%s' has not changed",
                upstream_full_ref_name,
            )



from __future__ import annotations

import os
from enum import Enum
from re import compile as regexp
from typing import TYPE_CHECKING

from semantic_release.globals import logger
from semantic_release.version.version import Version

if TYPE_CHECKING:
    from typing import Any

    from semantic_release.hvcs.github import Github


class PersistenceMode(Enum):
    TEMPORARY = "temporary"
    PERMANENT = "permanent"


class VersionGitHubActionsOutput:
    OUTPUT_ENV_VAR = "GITHUB_OUTPUT"

    def __init__(
        self,
        gh_client: Github | None = None,
        mode: PersistenceMode = PersistenceMode.PERMANENT,
        released: bool | None = None,
        version: Version | None = None,
        commit_sha: str | None = None,
        release_notes: str | None = None,
        prev_version: Version | None = None,
    ) -> None:
        self._gh_client = gh_client
        self._mode = mode
        self._released = released
        self._version = version
        self._commit_sha = commit_sha
        self._release_notes = release_notes
        self._prev_version = prev_version

    @property
    def released(self) -> bool | None:
        return self._released

    @released.setter
    def released(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise TypeError("output 'released' is boolean")
        self._released = value

    @property
    def version(self) -> Version | None:
        return self._version if self._version is not None else None

    @version.setter
    def version(self, value: Version) -> None:
        if not isinstance(value, Version):
            raise TypeError("output 'released' should be a Version")
        self._version = value

    @property
    def tag(self) -> str | None:
        return self.version.as_tag() if self.version is not None else None

    @property
    def is_prerelease(self) -> bool | None:
        return self.version.is_prerelease if self.version is not None else None

    @property
    def commit_sha(self) -> str | None:
        return self._commit_sha if self._commit_sha else None

    @commit_sha.setter
    def commit_sha(self, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError("output 'commit_sha' should be a string")

        if not regexp(r"^[0-9a-f]{40}$").match(value):
            raise ValueError(
                "output 'commit_sha' should be a valid 40-hex-character SHA"
            )

        self._commit_sha = value

    @property
    def release_notes(self) -> str | None:
        return self._release_notes if self._release_notes else None

    @release_notes.setter
    def release_notes(self, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError("output 'release_notes' should be a string")
        self._release_notes = value

    @property
    def prev_version(self) -> Version | None:
        if not self.released:
            return self.version
        return self._prev_version if self._prev_version else None

    @prev_version.setter
    def prev_version(self, value: Version) -> None:
        if not isinstance(value, Version):
            raise TypeError("output 'prev_version' should be a Version")
        self._prev_version = value

    @property
    def gh_client(self) -> Github:
        if not self._gh_client:
            raise ValueError("GitHub client not set, cannot create links")
        return self._gh_client

    def to_output_text(self) -> str:
        missing: set[str] = set()
        if self.version is None:
            missing.add("version")
        if self.released is None:
            missing.add("released")
        if self.released:
            if self.release_notes is None:
                missing.add("release_notes")
            if self._mode is PersistenceMode.PERMANENT and self.commit_sha is None:
                missing.add("commit_sha")

        if missing:
            raise ValueError(
                f"some required outputs were not set: {', '.join(missing)}"
            )

        output_values: dict[str, Any] = {
            "released": str(self.released).lower(),
            "version": str(self.version),
            "tag": self.tag,
            "is_prerelease": str(self.is_prerelease).lower(),
            "link": self.gh_client.create_release_url(self.tag) if self.tag else "",
            "previous_version": str(self.prev_version) if self.prev_version else "",
            "commit_sha": self.commit_sha if self.commit_sha else "",
        }

        multiline_output_values: dict[str, str] = {
            "release_notes": self.release_notes if self.release_notes else "",
        }

        output_lines = [
            *[f"{key}={value!s}{os.linesep}" for key, value in output_values.items()],
            *[
                f"{key}<<EOF{os.linesep}{value}EOF{os.linesep}"
                if value
                else f"{key}={os.linesep}"
                for key, value in multiline_output_values.items()
            ],
        ]

        return str.join("", output_lines)

    def write_if_possible(self, filename: str | None = None) -> None:
        output_file = filename or os.getenv(self.OUTPUT_ENV_VAR)
        if not output_file:
            logger.info("not writing GitHub Actions output, as no file specified")
            return

        with open(output_file, "ab") as f:
            f.write(self.to_output_text().encode("utf-8"))



from __future__ import annotations

from enum import Enum


class VersionStampType(str, Enum):
    """Enum for the type of version declaration"""

    # The version is a number format, e.g. 1.2.3
    NUMBER_FORMAT = "nf"

    TAG_FORMAT = "tf"



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pytest

# NOTE: use backport with newer API
from importlib_resources import files

import semantic_release
from semantic_release.changelog.context import ChangelogMode, make_changelog_context
from semantic_release.cli.changelog_writer import render_default_changelog_file
from semantic_release.cli.config import ChangelogOutputFormat
from semantic_release.commit_parser import ParsedCommit
from semantic_release.hvcs import Bitbucket, Gitea, Github, Gitlab

if TYPE_CHECKING:
    from semantic_release.changelog.release_history import ReleaseHistory


@pytest.fixture(scope="module")
def default_changelog_template() -> str:
    """Retrieve the semantic-release default changelog template."""
    version_notes_template = files(semantic_release.__name__).joinpath(
        Path("data", "templates", "conventional", "md", "CHANGELOG.md.j2")
    )
    return version_notes_template.read_text(encoding="utf-8")


@pytest.mark.parametrize("hvcs_client", [Github, Gitlab, Gitea, Bitbucket])
def test_default_changelog_template(
    hvcs_client: type[Bitbucket | Gitea | Github | Gitlab],
    example_git_https_url: str,
    artificial_release_history: ReleaseHistory,
    changelog_md_file: Path,
    today_date_str: str,
):
    artificial_release_history.unreleased = {}  # Wipe out unreleased
    hvcs = hvcs_client(example_git_https_url)

    latest_version = next(iter(artificial_release_history.released.keys()))
    latest_release = artificial_release_history.released[latest_version]

    first_version = list(artificial_release_history.released.keys())[-1]

    feat_commit_obj = latest_release["elements"]["feature"][0]
    fix_commit_obj_1 = latest_release["elements"]["fix"][0]
    fix_commit_obj_2 = latest_release["elements"]["fix"][1]
    fix_commit_obj_3 = latest_release["elements"]["fix"][2]
    assert isinstance(feat_commit_obj, ParsedCommit)
    assert isinstance(fix_commit_obj_1, ParsedCommit)
    assert isinstance(fix_commit_obj_2, ParsedCommit)
    assert isinstance(fix_commit_obj_3, ParsedCommit)

    feat_commit_url = hvcs.commit_hash_url(feat_commit_obj.commit.hexsha)
    feat_description = str.join("\n", feat_commit_obj.descriptions)

    fix_commit_1_url = hvcs.commit_hash_url(fix_commit_obj_1.commit.hexsha)
    fix_commit_1_description = str.join("\n", fix_commit_obj_1.descriptions)

    fix_commit_2_url = hvcs.commit_hash_url(fix_commit_obj_2.commit.hexsha)
    fix_commit_2_description = str.join("\n", fix_commit_obj_2.descriptions)

    fix_commit_3_url = hvcs.commit_hash_url(fix_commit_obj_3.commit.hexsha)
    fix_commit_3_description = str.join("\n", fix_commit_obj_3.descriptions)

    expected_changelog = str.join(
        "\n",
        [
            "# CHANGELOG",
            "",
            "",
            f"## v{latest_version} ({today_date_str})",
            "",
            "### Feature",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{feat_commit_obj.scope}**: {feat_description.capitalize()}",
            f"  ([`{feat_commit_obj.commit.hexsha[:7]}`]({feat_commit_url}))",
            "",
            "### Fix",
            "",
            # Commit 2 is first because it has no scope
            # Due to the 100 character limit, hash url will be on the second line
            f"- {fix_commit_2_description.capitalize()}",
            f"  ([`{fix_commit_obj_2.commit.hexsha[:7]}`]({fix_commit_2_url}))",
            "",
            # Commit 3 is second because it starts with an A even though it has the same scope as 1
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{fix_commit_obj_3.scope}**: {fix_commit_3_description.capitalize()}",
            f"  ([`{fix_commit_obj_3.commit.hexsha[:7]}`]({fix_commit_3_url}))",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{fix_commit_obj_1.scope}**: {fix_commit_1_description.capitalize()}",
            f"  ([`{fix_commit_obj_1.commit.hexsha[:7]}`]({fix_commit_1_url}))",
            "",
            "",
            f"## v{first_version} ({today_date_str})",
            "",
            "- Initial Release",
        ],
    )

    actual_changelog = render_default_changelog_file(
        output_format=ChangelogOutputFormat.MARKDOWN,
        changelog_context=make_changelog_context(
            hvcs_client=hvcs,
            release_history=artificial_release_history,
            mode=ChangelogMode.INIT,
            prev_changelog_file=changelog_md_file,
            insertion_flag="",
            mask_initial_release=True,
        ),
        changelog_style="conventional",
    )

    assert expected_changelog == actual_changelog


@pytest.mark.parametrize("hvcs_client", [Github, Gitlab, Gitea, Bitbucket])
def test_default_changelog_template_w_a_brk_change(
    hvcs_client: type[Bitbucket | Gitea | Github | Gitlab],
    example_git_https_url: str,
    release_history_w_brk_change: ReleaseHistory,
    changelog_md_file: Path,
    today_date_str: str,
):
    hvcs = hvcs_client(example_git_https_url)

    releases = iter(release_history_w_brk_change.released.keys())
    latest_version = next(releases)
    latest_release = release_history_w_brk_change.released[latest_version]

    previous_version = next(releases)
    previous_release = release_history_w_brk_change.released[previous_version]

    first_version = list(release_history_w_brk_change.released.keys())[-1]

    brk_fix_commit_obj = latest_release["elements"]["Bug Fixes"][0]
    feat_commit_obj = previous_release["elements"]["feature"][0]
    fix_commit_obj_1 = previous_release["elements"]["fix"][0]
    fix_commit_obj_2 = previous_release["elements"]["fix"][1]
    fix_commit_obj_3 = previous_release["elements"]["fix"][2]
    assert isinstance(brk_fix_commit_obj, ParsedCommit)
    assert isinstance(feat_commit_obj, ParsedCommit)
    assert isinstance(fix_commit_obj_1, ParsedCommit)
    assert isinstance(fix_commit_obj_2, ParsedCommit)
    assert isinstance(fix_commit_obj_3, ParsedCommit)

    brk_fix_commit_url = hvcs.commit_hash_url(brk_fix_commit_obj.commit.hexsha)
    brk_fix_description = str.join("\n", brk_fix_commit_obj.descriptions)
    brk_fix_brking_description = str.join(
        "\n", brk_fix_commit_obj.breaking_descriptions
    )

    feat_commit_url = hvcs.commit_hash_url(feat_commit_obj.commit.hexsha)
    feat_description = str.join("\n", feat_commit_obj.descriptions)

    fix_commit_1_url = hvcs.commit_hash_url(fix_commit_obj_1.commit.hexsha)
    fix_commit_1_description = str.join("\n", fix_commit_obj_1.descriptions)

    fix_commit_2_url = hvcs.commit_hash_url(fix_commit_obj_2.commit.hexsha)
    fix_commit_2_description = str.join("\n", fix_commit_obj_2.descriptions)

    fix_commit_3_url = hvcs.commit_hash_url(fix_commit_obj_3.commit.hexsha)
    fix_commit_3_description = str.join("\n", fix_commit_obj_3.descriptions)

    expected_changelog = str.join(
        "\n",
        [
            "# CHANGELOG",
            "",
            "",
            f"## v{latest_version} ({today_date_str})",
            "",
            "### Bug Fixes",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{brk_fix_commit_obj.scope}**: {brk_fix_description.capitalize()}",
            f"  ([`{brk_fix_commit_obj.commit.hexsha[:7]}`]({brk_fix_commit_url}))",
            "",
            "### Breaking Changes",
            "",
            # Currently does not consider the 100 character limit because the current
            # descriptions are short enough to fit in one line
            "- {commit_scope}{change_desc}".format(
                commit_scope=(
                    f"**{brk_fix_commit_obj.scope}**: "
                    if brk_fix_commit_obj.scope
                    else ""
                ),
                change_desc=brk_fix_brking_description.capitalize(),
            ),
            "",
            "",
            f"## v{previous_version} ({today_date_str})",
            "",
            "### Feature",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{feat_commit_obj.scope}**: {feat_description.capitalize()}",
            f"  ([`{feat_commit_obj.commit.hexsha[:7]}`]({feat_commit_url}))",
            "",
            "### Fix",
            "",
            # Commit 2 is first because it has no scope
            # Due to the 100 character limit, hash url will be on the second line
            f"- {fix_commit_2_description.capitalize()}",
            f"  ([`{fix_commit_obj_2.commit.hexsha[:7]}`]({fix_commit_2_url}))",
            "",
            # Commit 3 is second because it starts with an A even though it has the same scope as 1
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{fix_commit_obj_3.scope}**: {fix_commit_3_description.capitalize()}",
            f"  ([`{fix_commit_obj_3.commit.hexsha[:7]}`]({fix_commit_3_url}))",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{fix_commit_obj_1.scope}**: {fix_commit_1_description.capitalize()}",
            f"  ([`{fix_commit_obj_1.commit.hexsha[:7]}`]({fix_commit_1_url}))",
            "",
            "",
            f"## v{first_version} ({today_date_str})",
            "",
            "- Initial Release",
        ],
    )

    actual_changelog = render_default_changelog_file(
        output_format=ChangelogOutputFormat.MARKDOWN,
        changelog_context=make_changelog_context(
            hvcs_client=hvcs,
            release_history=release_history_w_brk_change,
            mode=ChangelogMode.INIT,
            prev_changelog_file=changelog_md_file,
            insertion_flag="",
            mask_initial_release=True,
        ),
        changelog_style="conventional",
    )

    assert expected_changelog == actual_changelog


@pytest.mark.parametrize("hvcs_client", [Github, Gitlab, Gitea, Bitbucket])
def test_default_changelog_template_w_multiple_brk_changes(
    hvcs_client: type[Bitbucket | Gitea | Github | Gitlab],
    example_git_https_url: str,
    release_history_w_multiple_brk_changes: ReleaseHistory,
    changelog_md_file: Path,
    today_date_str: str,
):
    hvcs = hvcs_client(example_git_https_url)

    releases = iter(release_history_w_multiple_brk_changes.released.keys())
    latest_version = next(releases)
    latest_release = release_history_w_multiple_brk_changes.released[latest_version]

    previous_version = next(releases)
    previous_release = release_history_w_multiple_brk_changes.released[previous_version]

    first_version = list(release_history_w_multiple_brk_changes.released.keys())[-1]

    brk_feat_commit_obj = latest_release["elements"]["Features"][0]
    brk_fix_commit_obj = latest_release["elements"]["Bug Fixes"][0]
    feat_commit_obj = previous_release["elements"]["feature"][0]
    fix_commit_obj_1 = previous_release["elements"]["fix"][0]
    fix_commit_obj_2 = previous_release["elements"]["fix"][1]
    fix_commit_obj_3 = previous_release["elements"]["fix"][2]
    assert isinstance(brk_feat_commit_obj, ParsedCommit)
    assert isinstance(brk_fix_commit_obj, ParsedCommit)
    assert isinstance(feat_commit_obj, ParsedCommit)
    assert isinstance(fix_commit_obj_1, ParsedCommit)
    assert isinstance(fix_commit_obj_2, ParsedCommit)
    assert isinstance(fix_commit_obj_3, ParsedCommit)

    brk_feat_commit_url = hvcs.commit_hash_url(brk_feat_commit_obj.commit.hexsha)
    brk_feat_description = str.join("\n", brk_feat_commit_obj.descriptions)
    brk_feat_brking_description = str.join(
        "\n", brk_feat_commit_obj.breaking_descriptions
    )

    brk_fix_commit_url = hvcs.commit_hash_url(brk_fix_commit_obj.commit.hexsha)
    brk_fix_description = str.join("\n", brk_fix_commit_obj.descriptions)
    brk_fix_brking_description = str.join(
        "\n", brk_fix_commit_obj.breaking_descriptions
    )

    feat_commit_url = hvcs.commit_hash_url(feat_commit_obj.commit.hexsha)
    feat_description = str.join("\n", feat_commit_obj.descriptions)

    fix_commit_1_url = hvcs.commit_hash_url(fix_commit_obj_1.commit.hexsha)
    fix_commit_1_description = str.join("\n", fix_commit_obj_1.descriptions)

    fix_commit_2_url = hvcs.commit_hash_url(fix_commit_obj_2.commit.hexsha)
    fix_commit_2_description = str.join("\n", fix_commit_obj_2.descriptions)

    fix_commit_3_url = hvcs.commit_hash_url(fix_commit_obj_3.commit.hexsha)
    fix_commit_3_description = str.join("\n", fix_commit_obj_3.descriptions)

    expected_changelog = str.join(
        "\n",
        [
            "# CHANGELOG",
            "",
            "",
            f"## v{latest_version} ({today_date_str})",
            "",
            "### Bug Fixes",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{brk_fix_commit_obj.scope}**: {brk_fix_description.capitalize()}",
            f"  ([`{brk_fix_commit_obj.commit.hexsha[:7]}`]({brk_fix_commit_url}))",
            "",
            "### Features",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- {brk_feat_description.capitalize()}",
            f"  ([`{brk_feat_commit_obj.commit.hexsha[:7]}`]({brk_feat_commit_url}))",
            "",
            "### Breaking Changes",
            "",
            # Currently does not consider the 100 character limit because the current
            # descriptions are short enough to fit in one line
            "- {commit_scope}{change_desc}".format(
                commit_scope=(
                    f"**{brk_feat_commit_obj.scope}**: "
                    if brk_feat_commit_obj.scope
                    else ""
                ),
                change_desc=brk_feat_brking_description.capitalize(),
            ),
            "",
            "- {commit_scope}{change_desc}".format(
                commit_scope=(
                    f"**{brk_fix_commit_obj.scope}**: "
                    if brk_fix_commit_obj.scope
                    else ""
                ),
                change_desc=brk_fix_brking_description.capitalize(),
            ),
            "",
            "",
            f"## v{previous_version} ({today_date_str})",
            "",
            "### Feature",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{feat_commit_obj.scope}**: {feat_description.capitalize()}",
            f"  ([`{feat_commit_obj.commit.hexsha[:7]}`]({feat_commit_url}))",
            "",
            "### Fix",
            "",
            # Commit 2 is first because it has no scope
            # Due to the 100 character limit, hash url will be on the second line
            f"- {fix_commit_2_description.capitalize()}",
            f"  ([`{fix_commit_obj_2.commit.hexsha[:7]}`]({fix_commit_2_url}))",
            "",
            # Commit 3 is second because it starts with an A even though it has the same scope as 1
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{fix_commit_obj_3.scope}**: {fix_commit_3_description.capitalize()}",
            f"  ([`{fix_commit_obj_3.commit.hexsha[:7]}`]({fix_commit_3_url}))",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{fix_commit_obj_1.scope}**: {fix_commit_1_description.capitalize()}",
            f"  ([`{fix_commit_obj_1.commit.hexsha[:7]}`]({fix_commit_1_url}))",
            "",
            "",
            f"## v{first_version} ({today_date_str})",
            "",
            "- Initial Release",
        ],
    )

    actual_changelog = render_default_changelog_file(
        output_format=ChangelogOutputFormat.MARKDOWN,
        changelog_context=make_changelog_context(
            hvcs_client=hvcs,
            release_history=release_history_w_multiple_brk_changes,
            mode=ChangelogMode.INIT,
            prev_changelog_file=changelog_md_file,
            insertion_flag="",
            mask_initial_release=True,
        ),
        changelog_style="conventional",
    )

    assert expected_changelog == actual_changelog


@pytest.mark.parametrize("hvcs_client", [Github, Gitlab, Gitea, Bitbucket])
def test_default_changelog_template_no_initial_release_mask(
    hvcs_client: type[Bitbucket | Gitea | Github | Gitlab],
    example_git_https_url: str,
    artificial_release_history: ReleaseHistory,
    changelog_md_file: Path,
    today_date_str: str,
):
    artificial_release_history.unreleased = {}  # Wipe out unreleased
    hvcs = hvcs_client(example_git_https_url)

    latest_version = next(iter(artificial_release_history.released.keys()))
    latest_release = artificial_release_history.released[latest_version]

    first_version = list(artificial_release_history.released.keys())[-1]

    feat_commit_obj = latest_release["elements"]["feature"][0]
    fix_commit_obj_1 = latest_release["elements"]["fix"][0]
    fix_commit_obj_2 = latest_release["elements"]["fix"][1]
    fix_commit_obj_3 = latest_release["elements"]["fix"][2]
    assert isinstance(feat_commit_obj, ParsedCommit)
    assert isinstance(fix_commit_obj_1, ParsedCommit)
    assert isinstance(fix_commit_obj_2, ParsedCommit)
    assert isinstance(fix_commit_obj_3, ParsedCommit)

    feat_commit_url = hvcs.commit_hash_url(feat_commit_obj.commit.hexsha)
    feat_description = str.join("\n", feat_commit_obj.descriptions)

    fix_commit_1_url = hvcs.commit_hash_url(fix_commit_obj_1.commit.hexsha)
    fix_commit_1_description = str.join("\n", fix_commit_obj_1.descriptions)

    fix_commit_2_url = hvcs.commit_hash_url(fix_commit_obj_2.commit.hexsha)
    fix_commit_2_description = str.join("\n", fix_commit_obj_2.descriptions)

    fix_commit_3_url = hvcs.commit_hash_url(fix_commit_obj_3.commit.hexsha)
    fix_commit_3_description = str.join("\n", fix_commit_obj_3.descriptions)

    expected_changelog = str.join(
        "\n",
        [
            "# CHANGELOG",
            "",
            "",
            f"## v{latest_version} ({today_date_str})",
            "",
            "### Feature",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{feat_commit_obj.scope}**: {feat_description.capitalize()}",
            f"  ([`{feat_commit_obj.commit.hexsha[:7]}`]({feat_commit_url}))",
            "",
            "### Fix",
            "",
            # Commit 2 is first because it has no scope
            # Due to the 100 character limit, hash url will be on the second line
            f"- {fix_commit_2_description.capitalize()}",
            f"  ([`{fix_commit_obj_2.commit.hexsha[:7]}`]({fix_commit_2_url}))",
            "",
            # Commit 3 is second because it starts with an A even though it has the same scope as 1
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{fix_commit_obj_3.scope}**: {fix_commit_3_description.capitalize()}",
            f"  ([`{fix_commit_obj_3.commit.hexsha[:7]}`]({fix_commit_3_url}))",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{fix_commit_obj_1.scope}**: {fix_commit_1_description.capitalize()}",
            f"  ([`{fix_commit_obj_1.commit.hexsha[:7]}`]({fix_commit_1_url}))",
            "",
            "",
            f"## v{first_version} ({today_date_str})",
            "",
            "### Feature",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{feat_commit_obj.scope}**: {feat_description.capitalize()}",
            f"  ([`{feat_commit_obj.commit.hexsha[:7]}`]({feat_commit_url}))",
        ],
    )

    actual_changelog = render_default_changelog_file(
        output_format=ChangelogOutputFormat.MARKDOWN,
        changelog_context=make_changelog_context(
            hvcs_client=hvcs,
            release_history=artificial_release_history,
            mode=ChangelogMode.INIT,
            prev_changelog_file=changelog_md_file,
            insertion_flag="",
            mask_initial_release=False,
        ),
        changelog_style="conventional",
    )

    assert expected_changelog == actual_changelog


@pytest.mark.parametrize("hvcs_client", [Github, Gitlab, Gitea, Bitbucket])
def test_default_changelog_template_w_unreleased_changes(
    hvcs_client: type[Bitbucket | Gitea | Github | Gitlab],
    example_git_https_url: str,
    artificial_release_history: ReleaseHistory,
    changelog_md_file: Path,
    today_date_str: str,
):
    hvcs = hvcs_client(example_git_https_url)

    latest_version = next(iter(artificial_release_history.released.keys()))
    latest_release = artificial_release_history.released[latest_version]

    first_version = list(artificial_release_history.released.keys())[-1]

    feat_commit_obj = latest_release["elements"]["feature"][0]
    fix_commit_obj_1 = latest_release["elements"]["fix"][0]
    fix_commit_obj_2 = latest_release["elements"]["fix"][1]
    fix_commit_obj_3 = latest_release["elements"]["fix"][2]
    assert isinstance(feat_commit_obj, ParsedCommit)
    assert isinstance(fix_commit_obj_1, ParsedCommit)
    assert isinstance(fix_commit_obj_2, ParsedCommit)
    assert isinstance(fix_commit_obj_3, ParsedCommit)

    feat_commit_url = hvcs.commit_hash_url(feat_commit_obj.commit.hexsha)
    feat_description = str.join("\n", feat_commit_obj.descriptions)

    fix_commit_1_url = hvcs.commit_hash_url(fix_commit_obj_1.commit.hexsha)
    fix_commit_1_description = str.join("\n", fix_commit_obj_1.descriptions)

    fix_commit_2_url = hvcs.commit_hash_url(fix_commit_obj_2.commit.hexsha)
    fix_commit_2_description = str.join("\n", fix_commit_obj_2.descriptions)

    fix_commit_3_url = hvcs.commit_hash_url(fix_commit_obj_3.commit.hexsha)
    fix_commit_3_description = str.join("\n", fix_commit_obj_3.descriptions)

    expected_changelog = str.join(
        "\n",
        [
            "# CHANGELOG",
            "",
            "",
            "## Unreleased",
            "",
            "### Feature",
            "",
            f"- **{feat_commit_obj.scope}**: {feat_description.capitalize()}",
            f"  ([`{feat_commit_obj.commit.hexsha[:7]}`]({feat_commit_url}))",
            "",
            "",
            f"## v{latest_version} ({today_date_str})",
            "",
            "### Feature",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{feat_commit_obj.scope}**: {feat_description.capitalize()}",
            f"  ([`{feat_commit_obj.commit.hexsha[:7]}`]({feat_commit_url}))",
            "",
            "### Fix",
            "",
            # Commit 2 is first because it has no scope
            # Due to the 100 character limit, hash url will be on the second line
            f"- {fix_commit_2_description.capitalize()}",
            f"  ([`{fix_commit_obj_2.commit.hexsha[:7]}`]({fix_commit_2_url}))",
            "",
            # Commit 3 is second because it starts with an A even though it has the same scope as 1
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{fix_commit_obj_3.scope}**: {fix_commit_3_description.capitalize()}",
            f"  ([`{fix_commit_obj_3.commit.hexsha[:7]}`]({fix_commit_3_url}))",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{fix_commit_obj_1.scope}**: {fix_commit_1_description.capitalize()}",
            f"  ([`{fix_commit_obj_1.commit.hexsha[:7]}`]({fix_commit_1_url}))",
            "",
            "",
            f"## v{first_version} ({today_date_str})",
            "",
            "- Initial Release",
        ],
    )

    actual_changelog = render_default_changelog_file(
        output_format=ChangelogOutputFormat.MARKDOWN,
        changelog_context=make_changelog_context(
            hvcs_client=hvcs,
            release_history=artificial_release_history,
            mode=ChangelogMode.INIT,
            prev_changelog_file=changelog_md_file,
            insertion_flag="",
            mask_initial_release=True,
        ),
        changelog_style="conventional",
    )

    assert expected_changelog == actual_changelog


@pytest.mark.parametrize("hvcs_client", [Github, Gitlab, Gitea, Bitbucket])
def test_default_changelog_template_w_a_notice(
    hvcs_client: type[Github | Gitlab | Gitea | Bitbucket],
    example_git_https_url: str,
    release_history_w_a_notice: ReleaseHistory,
    changelog_md_file: Path,
    today_date_str: str,
):
    hvcs = hvcs_client(example_git_https_url)
    released_versions = iter(release_history_w_a_notice.released.keys())

    latest_version = next(released_versions)
    prev_version_1 = next(released_versions)
    prev_version_2 = next(released_versions)

    latest_release = release_history_w_a_notice.released[latest_version]
    prev_release_1 = release_history_w_a_notice.released[prev_version_1]

    notice_commit_obj = next(iter(latest_release["elements"].values()))[0]
    feat_commit_obj = prev_release_1["elements"]["feature"][0]
    fix_commit_obj_1 = prev_release_1["elements"]["fix"][0]
    fix_commit_obj_2 = prev_release_1["elements"]["fix"][1]
    fix_commit_obj_3 = prev_release_1["elements"]["fix"][2]

    assert isinstance(notice_commit_obj, ParsedCommit)
    assert isinstance(feat_commit_obj, ParsedCommit)
    assert isinstance(fix_commit_obj_1, ParsedCommit)
    assert isinstance(fix_commit_obj_2, ParsedCommit)
    assert isinstance(fix_commit_obj_3, ParsedCommit)

    notice_commit_url = hvcs.commit_hash_url(notice_commit_obj.commit.hexsha)
    notice_commit_description = str.join("\n", notice_commit_obj.descriptions)
    notice_description = str.join("\n", notice_commit_obj.release_notices)

    feat_commit_url = hvcs.commit_hash_url(feat_commit_obj.commit.hexsha)
    feat_description = str.join("\n", feat_commit_obj.descriptions)

    fix_commit_1_url = hvcs.commit_hash_url(fix_commit_obj_1.commit.hexsha)
    fix_commit_1_description = str.join("\n", fix_commit_obj_1.descriptions)

    fix_commit_2_url = hvcs.commit_hash_url(fix_commit_obj_2.commit.hexsha)
    fix_commit_2_description = str.join("\n", fix_commit_obj_2.descriptions)

    fix_commit_3_url = hvcs.commit_hash_url(fix_commit_obj_3.commit.hexsha)
    fix_commit_3_description = str.join("\n", fix_commit_obj_3.descriptions)

    expected_changelog = str.join(
        "\n",
        [
            "# CHANGELOG",
            "",
            "",
            f"## v{latest_version} ({today_date_str})",
            "",
            "### Refactoring",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{notice_commit_obj.scope}**: {notice_commit_description.capitalize().rstrip()}",
            f"  ([`{notice_commit_obj.commit.hexsha[:7]}`]({notice_commit_url}))",
            "",
            "### Additional Release Information",
            "",
            "- {commit_scope}{change_desc}".format(
                commit_scope=(
                    f"**{notice_commit_obj.scope}**: "
                    if notice_commit_obj.scope
                    else ""
                ),
                change_desc=notice_description.capitalize().rstrip(),
            ),
            "",
            "",
            f"## v{prev_version_1} ({today_date_str})",
            "",
            "### Feature",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{feat_commit_obj.scope}**: {feat_description.capitalize()}",
            f"  ([`{feat_commit_obj.commit.hexsha[:7]}`]({feat_commit_url}))",
            "",
            "### Fix",
            "",
            # Commit 2 is first because it has no scope
            # Due to the 100 character limit, hash url will be on the second line
            f"- {fix_commit_2_description.capitalize()}",
            f"  ([`{fix_commit_obj_2.commit.hexsha[:7]}`]({fix_commit_2_url}))",
            "",
            # Commit 3 is second because it starts with an A even though it has the same scope as 1
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{fix_commit_obj_3.scope}**: {fix_commit_3_description.capitalize()}",
            f"  ([`{fix_commit_obj_3.commit.hexsha[:7]}`]({fix_commit_3_url}))",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{fix_commit_obj_1.scope}**: {fix_commit_1_description.capitalize()}",
            f"  ([`{fix_commit_obj_1.commit.hexsha[:7]}`]({fix_commit_1_url}))",
            "",
            "",
            f"## v{prev_version_2} ({today_date_str})",
            "",
            "### Feature",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{feat_commit_obj.scope}**: {feat_description.capitalize()}",
            f"  ([`{feat_commit_obj.commit.hexsha[:7]}`]({feat_commit_url}))",
        ],
    )

    actual_changelog = render_default_changelog_file(
        output_format=ChangelogOutputFormat.MARKDOWN,
        changelog_context=make_changelog_context(
            hvcs_client=hvcs,
            release_history=release_history_w_a_notice,
            mode=ChangelogMode.INIT,
            prev_changelog_file=changelog_md_file,
            insertion_flag="",
            mask_initial_release=False,
        ),
        changelog_style="conventional",
    )

    assert expected_changelog == actual_changelog


@pytest.mark.parametrize("hvcs_client", [Github, Gitlab, Gitea, Bitbucket])
def test_default_changelog_template_w_a_notice_n_brk_change(
    hvcs_client: type[Github | Gitlab | Gitea | Bitbucket],
    example_git_https_url: str,
    release_history_w_notice_n_brk_change: ReleaseHistory,
    changelog_md_file: Path,
    today_date_str: str,
):
    hvcs = hvcs_client(example_git_https_url)
    released_versions = iter(release_history_w_notice_n_brk_change.released.keys())

    latest_version = next(released_versions)
    prev_version_1 = next(released_versions)
    prev_version_2 = next(released_versions)

    latest_release = release_history_w_notice_n_brk_change.released[latest_version]
    prev_release_1 = release_history_w_notice_n_brk_change.released[prev_version_1]

    brk_fix_commit_obj = latest_release["elements"]["Bug Fixes"][0]
    notice_commit_obj = latest_release["elements"]["Refactoring"][0]
    feat_commit_obj = prev_release_1["elements"]["feature"][0]
    fix_commit_obj_1 = prev_release_1["elements"]["fix"][0]
    fix_commit_obj_2 = prev_release_1["elements"]["fix"][1]
    fix_commit_obj_3 = prev_release_1["elements"]["fix"][2]

    assert isinstance(brk_fix_commit_obj, ParsedCommit)
    assert isinstance(notice_commit_obj, ParsedCommit)
    assert isinstance(feat_commit_obj, ParsedCommit)
    assert isinstance(fix_commit_obj_1, ParsedCommit)
    assert isinstance(fix_commit_obj_2, ParsedCommit)
    assert isinstance(fix_commit_obj_3, ParsedCommit)

    brk_fix_commit_url = hvcs.commit_hash_url(brk_fix_commit_obj.commit.hexsha)
    brk_fix_description = str.join("\n", brk_fix_commit_obj.descriptions)
    brk_fix_brking_description = str.join(
        "\n", brk_fix_commit_obj.breaking_descriptions
    )

    notice_commit_url = hvcs.commit_hash_url(notice_commit_obj.commit.hexsha)
    notice_commit_description = str.join("\n", notice_commit_obj.descriptions)
    notice_description = str.join("\n", notice_commit_obj.release_notices)

    feat_commit_url = hvcs.commit_hash_url(feat_commit_obj.commit.hexsha)
    feat_description = str.join("\n", feat_commit_obj.descriptions)

    fix_commit_1_url = hvcs.commit_hash_url(fix_commit_obj_1.commit.hexsha)
    fix_commit_1_description = str.join("\n", fix_commit_obj_1.descriptions)

    fix_commit_2_url = hvcs.commit_hash_url(fix_commit_obj_2.commit.hexsha)
    fix_commit_2_description = str.join("\n", fix_commit_obj_2.descriptions)

    fix_commit_3_url = hvcs.commit_hash_url(fix_commit_obj_3.commit.hexsha)
    fix_commit_3_description = str.join("\n", fix_commit_obj_3.descriptions)

    expected_changelog = str.join(
        "\n",
        [
            "# CHANGELOG",
            "",
            "",
            f"## v{latest_version} ({today_date_str})",
            "",
            "### Bug Fixes",
            "",
            "- {commit_scope}{commit_desc}".format(
                commit_scope=(
                    f"**{brk_fix_commit_obj.scope}**: "
                    if brk_fix_commit_obj.scope
                    else ""
                ),
                commit_desc=brk_fix_description.capitalize().rstrip(),
            ),
            f"  ([`{brk_fix_commit_obj.commit.hexsha[:7]}`]({brk_fix_commit_url}))",
            "",
            "### Refactoring",
            "",
            "- {commit_scope}{commit_desc}".format(
                commit_scope=(
                    f"**{notice_commit_obj.scope}**: "
                    if notice_commit_obj.scope
                    else ""
                ),
                commit_desc=notice_commit_description.capitalize().rstrip(),
            ),
            f"  ([`{notice_commit_obj.commit.hexsha[:7]}`]({notice_commit_url}))",
            "",
            "### Breaking Changes",
            "",
            "- {commit_scope}{change_desc}".format(
                commit_scope=(
                    f"**{brk_fix_commit_obj.scope}**: "
                    if brk_fix_commit_obj.scope
                    else ""
                ),
                change_desc=brk_fix_brking_description.capitalize().rstrip(),
            ),
            "",
            "### Additional Release Information",
            "",
            "- {commit_scope}{change_desc}".format(
                commit_scope=(
                    f"**{notice_commit_obj.scope}**: "
                    if notice_commit_obj.scope
                    else ""
                ),
                change_desc=notice_description.capitalize().rstrip(),
            ),
            "",
            "",
            f"## v{prev_version_1} ({today_date_str})",
            "",
            "### Feature",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{feat_commit_obj.scope}**: {feat_description.capitalize()}",
            f"  ([`{feat_commit_obj.commit.hexsha[:7]}`]({feat_commit_url}))",
            "",
            "### Fix",
            "",
            # Commit 2 is first because it has no scope
            # Due to the 100 character limit, hash url will be on the second line
            f"- {fix_commit_2_description.capitalize()}",
            f"  ([`{fix_commit_obj_2.commit.hexsha[:7]}`]({fix_commit_2_url}))",
            "",
            # Commit 3 is second because it starts with an A even though it has the same scope as 1
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{fix_commit_obj_3.scope}**: {fix_commit_3_description.capitalize()}",
            f"  ([`{fix_commit_obj_3.commit.hexsha[:7]}`]({fix_commit_3_url}))",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{fix_commit_obj_1.scope}**: {fix_commit_1_description.capitalize()}",
            f"  ([`{fix_commit_obj_1.commit.hexsha[:7]}`]({fix_commit_1_url}))",
            "",
            "",
            f"## v{prev_version_2} ({today_date_str})",
            "",
            "### Feature",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{feat_commit_obj.scope}**: {feat_description.capitalize()}",
            f"  ([`{feat_commit_obj.commit.hexsha[:7]}`]({feat_commit_url}))",
        ],
    )

    actual_changelog = render_default_changelog_file(
        output_format=ChangelogOutputFormat.MARKDOWN,
        changelog_context=make_changelog_context(
            hvcs_client=hvcs,
            release_history=release_history_w_notice_n_brk_change,
            mode=ChangelogMode.INIT,
            prev_changelog_file=changelog_md_file,
            insertion_flag="",
            mask_initial_release=False,
        ),
        changelog_style="conventional",
    )

    assert expected_changelog == actual_changelog


@pytest.mark.parametrize("hvcs_client", [Github, Gitlab, Gitea, Bitbucket])
def test_default_changelog_template_w_multiple_notices(
    hvcs_client: type[Github | Gitlab | Gitea | Bitbucket],
    example_git_https_url: str,
    release_history_w_multiple_notices: ReleaseHistory,
    changelog_md_file: Path,
    today_date_str: str,
):
    hvcs = hvcs_client(example_git_https_url)
    released_versions = iter(release_history_w_multiple_notices.released.keys())

    latest_version = next(released_versions)
    prev_version_1 = next(released_versions)
    prev_version_2 = next(released_versions)

    latest_release = release_history_w_multiple_notices.released[latest_version]
    prev_release_1 = release_history_w_multiple_notices.released[prev_version_1]

    feat_notice_commit_obj = latest_release["elements"]["Features"][0]
    refactor_notice_commit_obj = latest_release["elements"]["Refactoring"][0]
    feat_commit_obj = prev_release_1["elements"]["feature"][0]
    fix_commit_obj_1 = prev_release_1["elements"]["fix"][0]
    fix_commit_obj_2 = prev_release_1["elements"]["fix"][1]
    fix_commit_obj_3 = prev_release_1["elements"]["fix"][2]

    assert isinstance(feat_notice_commit_obj, ParsedCommit)
    assert isinstance(refactor_notice_commit_obj, ParsedCommit)
    assert isinstance(feat_commit_obj, ParsedCommit)
    assert isinstance(fix_commit_obj_1, ParsedCommit)
    assert isinstance(fix_commit_obj_2, ParsedCommit)
    assert isinstance(fix_commit_obj_3, ParsedCommit)

    refactor_commit_url = hvcs.commit_hash_url(refactor_notice_commit_obj.commit.hexsha)
    refactor_commit_desc = str.join("\n", refactor_notice_commit_obj.descriptions)
    refactor_commit_notice_desc = str.join(
        "\n", refactor_notice_commit_obj.release_notices
    )

    feat_notice_commit_url = hvcs.commit_hash_url(feat_notice_commit_obj.commit.hexsha)
    feat_notice_description = str.join("\n", feat_notice_commit_obj.descriptions)
    feat_commit_notice_desc = str.join("\n", feat_notice_commit_obj.release_notices)

    feat_commit_url = hvcs.commit_hash_url(feat_commit_obj.commit.hexsha)
    feat_description = str.join("\n", feat_commit_obj.descriptions)

    fix_commit_1_url = hvcs.commit_hash_url(fix_commit_obj_1.commit.hexsha)
    fix_commit_1_description = str.join("\n", fix_commit_obj_1.descriptions)

    fix_commit_2_url = hvcs.commit_hash_url(fix_commit_obj_2.commit.hexsha)
    fix_commit_2_description = str.join("\n", fix_commit_obj_2.descriptions)

    fix_commit_3_url = hvcs.commit_hash_url(fix_commit_obj_3.commit.hexsha)
    fix_commit_3_description = str.join("\n", fix_commit_obj_3.descriptions)

    expected_changelog = str.join(
        "\n",
        [
            "# CHANGELOG",
            "",
            "",
            f"## v{latest_version} ({today_date_str})",
            "",
            "### Features",
            "",
            "- {commit_scope}{commit_desc}".format(
                commit_scope=(
                    f"**{feat_notice_commit_obj.scope}**: "
                    if feat_notice_commit_obj.scope
                    else ""
                ),
                commit_desc=feat_notice_description.capitalize().rstrip(),
            ),
            f"  ([`{feat_notice_commit_obj.commit.hexsha[:7]}`]({feat_notice_commit_url}))",
            "",
            "### Refactoring",
            "",
            "- {commit_scope}{commit_desc}".format(
                commit_scope=(
                    f"**{refactor_notice_commit_obj.scope}**: "
                    if refactor_notice_commit_obj.scope
                    else ""
                ),
                commit_desc=refactor_commit_desc.capitalize().rstrip(),
            ),
            f"  ([`{refactor_notice_commit_obj.commit.hexsha[:7]}`]({refactor_commit_url}))",
            "",
            "### Additional Release Information",
            "",
            "- {commit_scope}{change_desc}".format(
                commit_scope=(
                    f"**{refactor_notice_commit_obj.scope}**: "
                    if refactor_notice_commit_obj.scope
                    else ""
                ),
                change_desc=refactor_commit_notice_desc.capitalize().rstrip(),
            ),
            "",
            "- {commit_scope}{change_desc}".format(
                commit_scope=(
                    f"**{feat_notice_commit_obj.scope}**: "
                    if feat_notice_commit_obj.scope
                    else ""
                ),
                change_desc=str.join(
                    "\n",
                    [
                        feat_commit_notice_desc.capitalize()[:73].rstrip(),
                        "  " + feat_commit_notice_desc[73:].strip(),
                    ],
                ),
            ),
            "",
            "",
            f"## v{prev_version_1} ({today_date_str})",
            "",
            "### Feature",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{feat_commit_obj.scope}**: {feat_description.capitalize()}",
            f"  ([`{feat_commit_obj.commit.hexsha[:7]}`]({feat_commit_url}))",
            "",
            "### Fix",
            "",
            # Commit 2 is first because it has no scope
            # Due to the 100 character limit, hash url will be on the second line
            f"- {fix_commit_2_description.capitalize()}",
            f"  ([`{fix_commit_obj_2.commit.hexsha[:7]}`]({fix_commit_2_url}))",
            "",
            # Commit 3 is second because it starts with an A even though it has the same scope as 1
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{fix_commit_obj_3.scope}**: {fix_commit_3_description.capitalize()}",
            f"  ([`{fix_commit_obj_3.commit.hexsha[:7]}`]({fix_commit_3_url}))",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{fix_commit_obj_1.scope}**: {fix_commit_1_description.capitalize()}",
            f"  ([`{fix_commit_obj_1.commit.hexsha[:7]}`]({fix_commit_1_url}))",
            "",
            "",
            f"## v{prev_version_2} ({today_date_str})",
            "",
            "### Feature",
            "",
            # Due to the 100 character limit, hash url will be on the second line
            f"- **{feat_commit_obj.scope}**: {feat_description.capitalize()}",
            f"  ([`{feat_commit_obj.commit.hexsha[:7]}`]({feat_commit_url}))",
        ],
    )

    actual_changelog = render_default_changelog_file(
        output_format=ChangelogOutputFormat.MARKDOWN,
        changelog_context=make_changelog_context(
            hvcs_client=hvcs,
            release_history=release_history_w_multiple_notices,
            mode=ChangelogMode.INIT,
            prev_changelog_file=changelog_md_file,
            insertion_flag="",
            mask_initial_release=False,
        ),
        changelog_style="conventional",
    )

    assert expected_changelog == actual_changelog



import pytest

from semantic_release.const import SEMVER_REGEX
from semantic_release.version.translator import VersionTranslator
from semantic_release.version.version import Version

from tests.const import (
    A_FULL_VERSION_STRING,
    A_FULL_VERSION_STRING_WITH_BUILD_METADATA,
    A_PRERELEASE_VERSION_STRING,
)


@pytest.fixture
def a_full_version() -> Version:
    return Version.parse(A_FULL_VERSION_STRING)


@pytest.fixture
def a_prerelease_version() -> Version:
    return Version.parse(A_PRERELEASE_VERSION_STRING)


@pytest.fixture
def a_full_version_with_build_metadata() -> Version:
    return Version.parse(A_FULL_VERSION_STRING_WITH_BUILD_METADATA)


@pytest.mark.parametrize(
    "version_string",
    [
        A_FULL_VERSION_STRING,
        A_PRERELEASE_VERSION_STRING,
        A_FULL_VERSION_STRING_WITH_BUILD_METADATA,
        "3.2.3-alpha.dev3+local.12345",  # Pretty much as complex an example as there is
    ],
)
def test_succeeds_semver_regex_match(version_string: str):
    assert SEMVER_REGEX.fullmatch(
        version_string
    ), "a valid semantic version was not matched"


@pytest.mark.parametrize(
    "invalid_version_str",
    ["v1.2.3", "2.1", "3.1.1..3", "4.1.1.dev3"],  # PEP440 version
)
def test_invalid_semver_not_matched(invalid_version_str: str):
    assert SEMVER_REGEX.fullmatch(invalid_version_str) is None


@pytest.mark.parametrize("fmt", ["version", "{versioN}", "v{major}.{minor}.{patch}"])
def test_invalid_tag_format(fmt: str):
    with pytest.raises(ValueError) as err:
        VersionTranslator(tag_format=fmt)
    assert all(("tag_format" in str(err), "version" in str(err), fmt in str(err)))


@pytest.mark.parametrize(
    "version_string",
    [
        A_FULL_VERSION_STRING,
        A_PRERELEASE_VERSION_STRING,
        A_FULL_VERSION_STRING_WITH_BUILD_METADATA,
    ],
)
@pytest.mark.parametrize(
    "tag_format, prerelease_token",
    [
        ("v{version}", "dev"),
        ("v{version}", "alpha"),
        ("special-tagging-scheme-{version}", "rc"),
    ],
)
def test_translator_converts_versions_with_default_formatting_rules(
    version_string: str, tag_format: str, prerelease_token: str
):
    translator = VersionTranslator(
        tag_format=tag_format, prerelease_token=prerelease_token
    )

    expected_version_obj = Version.parse(
        version_string, prerelease_token=translator.prerelease_token
    )
    expected_tag = tag_format.format(version=version_string)
    actual_version_obj = translator.from_string(version_string)
    actual_tag = translator.str_to_tag(version_string)

    # These are important assumptions for formatting into source files/tags/etc
    assert version_string == str(actual_version_obj)
    assert expected_version_obj == actual_version_obj
    assert expected_tag == actual_tag
    assert expected_version_obj == (translator.from_tag(expected_tag) or "")
    assert version_string == str(translator.from_tag(actual_tag) or "")
